from typing import ClassVar

from gemd import ProcessTemplate

from openmsimodel.entity.base.process import Process
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
)

__all__ = ["SettingTraveler"]


class SettingTraveler(Process):
    """Class representing the set up of traveler"""

    TEMPLATE: ClassVar[ProcessTemplate] = ProcessTemplate(
        name="Setting traveler",
        description="""Setting traveler
                """,
    )

    _ATTRS: ClassVar[AttrsDict] = {"conditions": {}, "parameters": {}}

    finalize_template(_ATTRS, TEMPLATE)



from typing import ClassVar

from gemd import (
    MaterialTemplate,
    ParameterTemplate,
    CategoricalBounds,
    NominalCategorical,
)

from openmsimodel.entity.base.material import Material
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
    _validate_temp_keys
)

import pandas as pd

# __all__ = ["SummarySheet"]


class Summary(Material):
    """Class representing summary """

    TEMPLATE: ClassVar[MaterialTemplate] = MaterialTemplate(
        "Summary",
        description="Summary sheet of all the experiments conducted in previous batch that will help generate the next compositions spaces and suggest compositions via Bayesian Optimizization",
    )

    _ATTRS: ClassVar[AttrsDict] = _validate_temp_keys(TEMPLATE)

    # define_attribute(
    #     _ATTRS,
    #     template=ParameterTemplate(
    #         name='Iteration',
    #         bounds=CategoricalBounds(categories=['AAA, AAB, AAC'])
    #     ),
    #     default_value=NominalCategorical('')
    # )

    finalize_template(_ATTRS, TEMPLATE)



# """Base material, process, and measurement classes."""

# from . import base_element
# from . import material
# from . import process
# from . import measurement
# from . import ingredient


# from .base_element import *
# from .material import *
# from .process import *
# from .measurement import *
# from .ingredient import *

# # # from .typing import *
# # # from .impl import *

# __all__ = (
#     base_element.__all__
#     + material.__all__
#     + process.__all__
#     + measurement.__all__
#     + ingredient.__all__
#     # + typing.__all__
#     # + impl.__all__
# )

# del base_element
# del material
# del process
# del measurement
# del ingredient
# # del typing
# # del impl



'''Support for segment (ramp/hold) attributes.'''

from typing import TypeAlias, ClassVar, Union, Optional, Any
import re

import numpy as np
import pandas as pd

from gemd import (
    ProcessTemplate, ProcessSpec, ProcessRun, ParameterTemplate,
    Condition, Parameter, FileLink,
    NominalCategorical, EmpiricalFormula, NominalInteger, NominalReal
)
from gemd.entity.template.attribute_template import AttributeTemplate
from gemd.entity.attribute.base_attribute import BaseAttribute

from .base_special_attrs import BaseSpecialAttrs
from .utils import AttrTypes, attr_template_dict, generate_tags, generate_source

__all__ = ['SegAttrs']

Process: TypeAlias = Union[ProcessSpec, ProcessRun]

class SegAttrs(BaseSpecialAttrs):
    '''
    Represent a single process's segment (ramp/hold) attributes as
    ``DataFrame``s of attributes.

    Each row of the ``DataFrame``s represents a segment, in order. Each column
    is an attribute name. The special ``Segment type`` parameter must be one of
    the columns to indicate whether or not the segment is a ramp or hold
    segment.
    '''

    SEG: ClassVar[str] = '_seg'
    RE: ClassVar[re.Pattern] = re.compile(f'^(.+){SEG}(\\d+)$')
    SEG_PARAM_NAME: ClassVar[str] = 'Segment type'
    CAT_STR: ClassVar[str] = 'categorical'
    FORM_STR: ClassVar[str] = 'formula'

    def __init__(self, process: Process) -> None:
        '''
        Parse a process's segment attributes and create ``DataFrame``s for the
        conditions and parameters.

        Segment attributes are expected to be named as
        ``<base_name>_seg<segment>``.

        Parameters
        ----------
        process: ProcessSpec or ProcessRun
            Process containing appropriately-named segment attributes.
        '''        

        self._process = process

        conditions_tuple = self._parse(process.conditions)
        self._seg_conds: pd.DataFrame = conditions_tuple[0]
        self._other_conds: list[Condition] = conditions_tuple[1]

        parameters_tuple = self._parse(process.parameters)
        self._seg_params: pd.DataFrame = parameters_tuple[0]
        self._other_params: list[Parameter] = parameters_tuple[1]

        self._reindex_all_dfs()

    @property
    def process(self) -> Process:
        '''The process containing these attributes.'''
        return self._process

    @property
    def segment_conditions(self) -> pd.DataFrame:
        '''Conditions of ramp/hold segments.'''
        return self._seg_conds
    
    @property
    def other_conditions(self) -> list[Condition]:
        '''Non-segment conditions.'''
        return self._other_conds

    @property
    def segment_parameters(self) -> pd.DataFrame:
        '''Parameters of ramp/hold segments.'''
        return self._seg_params
    
    @property
    def other_parameters(self) -> list[Parameter]:
        '''Non-segment parameters.'''
        return self._other_params

    def concat(self, df: pd.DataFrame) -> None:
        '''
        Concatenate segment attribute data.

        Indices in the new data are respected and will replace rows with
        identical indices in the existing data.

        Parameters
        ----------
        df: DataFrame
            Should contain the appropriate attribute columns to update the
            segments.
        '''

        attrs = pd.concat(self._df_list(), axis=1)
        attrs = pd.concat((attrs, df), axis=0)
        attrs = attrs[~attrs.index.duplicated(keep='last')]

        self._seg_conds = attrs.loc[:, self._seg_conds.columns]
        self._seg_params = attrs.loc[:, self._seg_params.columns]

        self._reindex_all_dfs()
        self._reindex_all_attrs()

    def remove(self, index: int) -> None:
        '''Remove a row from the attributes by index, then reindex.'''

        new_index = np.arange(len(self._seg_conds)-1)

        for df in self._df_list():
            df.drop(index=index, inplace=True)
            df.index = new_index

        self._reindex_all_attrs()

    def update_object(self) -> None:
        '''
        Update the ``ProcessSpec`` or ``ProcessRun`` used to initialize
        with the attributes.

        This is performed in-place on the existing process and will
        overwrite any attributes added after ``SegAttrs`` initialization.
        '''

        seg_conds = self._extract_attrs_from_df(self._seg_conds)
        seg_params = self._extract_attrs_from_df(self._seg_params)

        self._process.conditions = self._other_conds + seg_conds
        self._process.parameters = self._other_params + seg_params

    @classmethod
    def object_from_file(
        cls,
        filename: str,
        url: str,
        process_template: ProcessTemplate,
        attr_types: AttrTypes,
        columns: list[str],
        names: list[str],
        units: list[Union[str, None]],
        origins: list[str],
        brand: Optional[str] = None,
        model: Optional[str] = None,
        instr_id: Optional[str] = None,
        email: Optional[str] = None,
        iso_date: Optional[str] = None,
        **read_csv_kwargs: Any
        ) -> tuple[ProcessSpec, ProcessRun]:
        '''
        Create an empty ``ProcessSpec`` and filled ``ProcessRun`` from a delimited file.

        One column must be the special ``"Segment type"`` column. The values in
        this column are expected to be ``"ramp"`` or ``"hold"``, indicating the
        segment type.

        Parameters
        ----------
        filename: str
            The name of the file.
        url: str
            The URL or file path leading to `filename`.
        process_template: ProcessTemplate
            The process template for this process. Must contain a
            ``ParameterTemplate`` with name ``"Segment type"``.
        attr_types: list of Type[Condition], Type[Parameter], Type[Property]
            The type of each attribute in the data, in order. The element
            corresponding to ``"Segment type"`` must be ``Parameter``.
        columns: list[str]
            The names of the data columns.
        names: list[str]
            The names of the attributes corresponding to an attribute in
            `process_template`. One must be ``"Segment type"``.
        units: list[str or None]
            The units of the attribute values. ``"categorical"`` for
            NominalCategorical, ``"formula"``, for EmpiricalFormula,
            other ``str`` for NominalReal, ``None`` for NominalInteger.
        origins: list[str]
            The origins of the attribute values.
        brand: str, optional
            The brand of the instrument used for the process.
        model: str, optional
            The model of the instrument used for the process.
            Ignored if `brand` is ``None``.
        instr_id: str, optional
            A further identifier for the instrument used, such as a number.
            Ignored if `brand` or `model` is ``None``.
        email: str, optional
            Email of the person who performed the process.
        iso_date: str, optional
            Date the process was performed, in ISO format.
        **read_csv_kwargs: Any
            Passed to ``pandas.read_csv()``.
        '''

        df = pd.read_csv(filename, **read_csv_kwargs)

        tags = generate_tags(brand, model, instr_id)

        source = generate_source(email, iso_date)

        template_dict = attr_template_dict(process_template)

        if not (len(attr_types) == len(columns) == len(names) == len(units) == len(origins)):
            raise ValueError('Iterable inputs must be equal in length.')

        # check for segment parameter template in process template
        if (
            cls.SEG_PARAM_NAME not in template_dict
            or not isinstance(template_dict[cls.SEG_PARAM_NAME], ParameterTemplate)
        ):
            raise ValueError(
                f'process_template must contain a ParameterTemplate named "{cls.SEG_PARAM_NAME}"'
            )

        # check for segment parameter in iterables
        if (
            cls.SEG_PARAM_NAME not in names
            or attr_types[names.index(cls.SEG_PARAM_NAME)] is not Parameter
            or units[names.index(cls.SEG_PARAM_NAME)] != cls.CAT_STR
        ):
            raise ValueError(
                f'names must contain "{cls.SEG_PARAM_NAME}" with corresponding '
                f'Type[Parameter] in attr_types and "{cls.CAT_STR}" in units.'
            )

        conditions, parameters = cls._create_attrs_from_df(
            df, template_dict, attr_types, columns, names, units, origins
        )

        spec = ProcessSpec(
            name=process_template.name,
            template=process_template,
            tags=tags
        )

        run = ProcessRun(
            name=process_template.name,
            spec=spec,
            tags=tags,
            source=source,
            conditions=conditions,
            parameters=parameters,
            file_links=[FileLink(filename=filename, url=url)]
        )

        return spec, run

    @classmethod
    def _parse(cls, attrs: list[BaseAttribute]) -> tuple[pd.DataFrame, list[BaseAttribute]]:
        '''
        Match each attribute's name with regular expression.

        Base name determines the column. _seg value determines the row.
        '''

        seg_df = pd.DataFrame()
        other_attrs = []

        for attr in attrs:

            match = cls.RE.match(attr.name)

            if match is not None:
                cls._add_attr_to_df(attr, match, seg_df)
            else:
                other_attrs.append(attr)

        reindexed_df = cls._reindex_df(seg_df)

        return reindexed_df, other_attrs

    @classmethod
    def _add_attr_to_df(
        cls,
        attr: BaseAttribute,
        match: re.Match[str],
        seg_df: pd.DataFrame
        ) -> None:
        '''Given a matching attribute, add it to `seg_df`.'''

        # regular expression groups
        match_groups = match.groups()
        base_name = match_groups[0]
        seg_val = int(match_groups[1])
  
        # add column if not there
        if base_name not in seg_df.columns:
            seg_df[base_name] = np.nan

        # place attribute in appropriate df
        seg_df.loc[seg_val, base_name] = attr

    @classmethod
    def _reindex_df(cls, seg_df: pd.DataFrame) -> pd.DataFrame:
        '''Add in any missing integer indices and sort.'''

        if len(seg_df) > 0:
            seg_df = seg_df.reindex(np.arange(seg_df.index.max()+1))

        return seg_df

    def _reindex_all_dfs(self) -> None:
        '''Ensure that each ``DataFrame`` has the same (maximum) number of segments.'''

        if len(self._seg_conds) == len(self._seg_params):
            return

        seg_max = int(np.nanmax([self._seg_conds.index.max(), self._seg_params.index.max()]))

        self._seg_conds = self._seg_conds.reindex(np.arange(seg_max+1))
        self._seg_params = self._seg_params.reindex(np.arange(seg_max+1))

    def _df_list(self) -> list[pd.DataFrame]:
        '''Return a list of the attributes ``DataFrame``s.'''
        return [self._seg_conds, self._seg_params]

    def _reindex_all_attrs(self) -> None:
        '''Rename all attributes according to where they are in their ``DataFrame``.'''

        for df in self._df_list():
            for name, series in df.items():
                for i, attr in series.items():
                    if not pd.isna(attr):
                        attr.name = f'{name}{self.SEG}{i}'

    @classmethod
    def _create_attrs_from_df(
        cls,
        df: pd.DataFrame,
        template_dict: dict[str, AttributeTemplate],
        attr_types: AttrTypes,
        columns: list[str],
        names: list[str],
        units: list[Union[str, None]],
        origins: list[str]
        ) -> tuple[list[Condition], list[Parameter]]:
        '''Create conditions and parameters from the `df` values.'''

        conditions = []
        parameters = []

        for a, c, n, u, o in zip(attr_types, columns, names, units, origins):

            if a is Condition:
                to_append = conditions
            elif a is Parameter:
                to_append = parameters
            else:
                raise ValueError(
                    'attr_types must consist of Type[Condition] or Type[Parameter], '
                    f'but found {type(a)}.'
                )

            for i, val in enumerate(df.loc[:, c]):

                if u == cls.CAT_STR:
                    value = NominalCategorical(val)
                elif u == cls.FORM_STR:
                    value = EmpiricalFormula(val)
                elif u is None:
                    value = NominalInteger(val)
                else:
                    value = NominalReal(val, u)

                to_append.append(a(
                    name=f'{n}{cls.SEG}{i}',
                    value=value,
                    template=template_dict[n],
                    origin=o
                ))

        return conditions, parameters



from typing import ClassVar

from gemd import (
    ProcessTemplate,
    ParameterTemplate,
    CategoricalBounds,
    NominalCategorical,
)

from openmsimodel.entity.base.process import Process
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
)

__all__ = ["Mixing"]


class Mixing(Process):
    """Class representing the mixing of materials/elements for an experiment"""

    TEMPLATE: ClassVar[ProcessTemplate] = ProcessTemplate(
        name="Mixing",
        description="""Mixing individual elements of the composition to make the initial mixture for the alloy 
                """,
    )

    _ATTRS: ClassVar[AttrsDict] = {"conditions": {}, "parameters": {}}

    finalize_template(_ATTRS, TEMPLATE)



"""Classes for OpenMSIStream logging infrastructure"""

# imports
import pathlib, logging, traceback
from .has_arguments import HasArguments


class OpenMSIStreamFormatter(logging.Formatter):
    """
    Very small extension of the usual logging.Formatter to allow modification of format based on message content
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def format(self, record):
        """
        If a message starts with a newline, start the actual logging line with the newline before any of the rest
        """
        formatted = ""
        if record.msg.startswith("\n"):
            record.msg = record.msg.lstrip("\n")
            formatted += "\n"
        formatted += super().format(record)
        return formatted


class Logger:
    """
    Class for a general logger in OpenMSIStream format.

    :param logger_name: The name for the logger to use (automatically inferred from the running module if not given)
    :type logger_name: str, optional
    :param streamlevel: The level at/above which messages should be logged to the stream/console
    :type streamlevel: logging level int, optional
    :param logger_filepath: The path to a logger file to use or directory in which an automatically-named
        logger file should be created
    :type logger_filepath: :class:`pathlib.Path`, optional
    :param filelevel: The level at/above which messages should be written to the logfile
    :type filelevel: logging level int, optional
    """

    FORMATTER = OpenMSIStreamFormatter(
        "[%(name)s at %(asctime)s] %(message)s", "%Y-%m-%d %H:%M:%S"
    )

    def __init__(
        self,
        logger_name=None,
        streamlevel=logging.INFO,
        logger_filepath=None,
        filelevel=logging.WARNING,
    ):
        """
        name = the name for this logger to use (probably something like the top module that owns it)
        """
        self._name = logger_name
        if self._name is None:
            self._name = self.__class__.__name__
        self._logger_obj = logging.getLogger(self._name)
        self._logger_obj.setLevel(logging.DEBUG)
        self._streamhandler = logging.StreamHandler()
        self._streamhandler.setLevel(streamlevel)
        self._streamhandler.setFormatter(self.FORMATTER)
        self._logger_obj.addHandler(self._streamhandler)
        self._filehandler = None
        if logger_filepath is not None:
            self.add_file_handler(logger_filepath, level=filelevel)

    def set_level(self, level):
        """
        Set the level of the entire underlying logger

        :param level: The level to set
        :type level: logging level int
        """
        self._logger_obj.setLevel(level)

    def set_stream_level(self, level):
        """
        Set the level of the underlying logger's streamhandler

        :param level: The level to set
        :type level: logging level int
        """
        self._streamhandler.setLevel(level)

    def set_file_level(self, level):
        """
        Set the level of the underlying logger's filehandler

        :param level: The level to set
        :type level: logging level int
        """
        if self._filehandler is None:
            errmsg = f"ERROR: Logger {self._name} does not have a filehandler set but set_file_level was called!"
            raise RuntimeError(errmsg)
        self._filehandler.setLevel(level)

    def add_file_handler(self, filepath, level=logging.INFO):
        """
        Add an additional :class:`logging.FileHandler` to the logger

        :param filepath: The path to the new logger file
        :type filepath: :class:`pathlib.Path`
        :param level: The level to set
        :type level: logging level int
        """
        if not isinstance(filepath, pathlib.PurePath):
            self.error(
                f"ERROR: {filepath} is a {type(filepath)} object, not a Path object!",
                TypeError,
            )
        if not filepath.is_file():
            if not filepath.parent.is_dir():
                filepath.parent.mkdir(parents=True)
            filepath.touch()
        self._filehandler = logging.FileHandler(filepath)
        self._filehandler.setLevel(level)
        self._filehandler.setFormatter(self.FORMATTER)
        self._logger_obj.addHandler(self._filehandler)

    # methods for logging different levels of messages

    def debug(self, msg, *args, **kwargs):
        """
        Log a message at DEBUG level. Additional args/kwargs are sent to the underlying logger object's debug call.

        :param msg: the message to log
        :type msg: str
        """
        self._logger_obj.debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        """
        Log a message at INFO level. Additional args/kwargs are sent to the underlying logger object's info call.

        :param msg: the message to log
        :type msg: str
        """
        self._logger_obj.info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        """
        Log a message at WARNING level. Additional args/kwargs are sent to the underlying logger object's warning call.

        :param msg: the message to log (will have "WARNING: " prepended if it doesn't start with it)
        :type msg: str
        """
        if not msg.startswith("WARNING:"):
            msg = f"WARNING: {msg}"
        self._logger_obj.warning(msg, *args, **kwargs)

    # log an error message and optionally raise an exception with the same message, or raise a different exception
    def error(self, msg, exception_type=None, exc_obj=None, reraise=True, **kwargs):
        """
        Log a message at ERROR level. Optionally raise an exception of a given type with the same message.
        Optionally log the traceback of a given Exception at ERROR level, reraising it afterward if desired.

        Additional kwargs are sent to the underlying logger object's error call(s).

        :param msg: the message to log
        :type msg: str
        :param exception_type: The type of Exception to raise with the same message as `msg`
        :type exception_type: :class:`BaseException`, optional
        :param exc_obj: An Exception object whose traceback should be logged at error level
        :type exc_obj: :class:`BaseException`, optional
        :param reraise: if True, the given `exc_obj` will be reraised after it's logged.
        :type reraise: bool
        """
        if not msg.startswith("ERROR:"):
            msg = f"ERROR: {msg}"
        self._logger_obj.error(msg, **kwargs)
        if exc_obj is not None:
            self.log_exception_as_error(exc_obj, reraise=reraise, **kwargs)
        if exception_type is not None:
            raise exception_type(msg)

    def log_exception_as_error(self, exc, reraise=True, **kwargs):
        """
        Log the traceback of a given Exception as an error and optionally reraise it

        Additional kwargs are sent to the underlying logger object's error call.

        :param exc: An Exception object whose traceback should be logged at error level
        :type exc: :class:`BaseException`, optional
        :param reraise: if True, the given `exc` will be reraised after it's logged.
        :type reraise: bool
        """
        try:
            raise exc
        except Exception:
            exc_to_log = "ERROR: Exception: ["
            for line in (traceback.format_exc()).split("\n"):
                exc_to_log += f"{line},"
            exc_to_log += "]"
            self._logger_obj.error(exc_to_log, **kwargs)
            if reraise:
                raise exc



from typing import ClassVar

from gemd import ProcessTemplate, ParameterTemplate, RealBounds

from openmsimodel.entity.base.process import Process
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
)

__all__ = ["Forging"]


class Forging(Process):
    """Class representing forging cycle"""

    TEMPLATE: ClassVar[ProcessTemplate] = ProcessTemplate(
        name="Forging",
        description="""Forging
                """,
    )

    _ATTRS: ClassVar[AttrsDict] = {"conditions": {}, "parameters": {}}

    define_attribute(
        _ATTRS,
        template=ParameterTemplate(
            name="Soak Time",
            bounds=RealBounds(0, 120, "minutes"),
        ),
    )

    define_attribute(
        _ATTRS,
        template=ParameterTemplate(
            name="Temperature",
            bounds=RealBounds(0, 2000, "Kelvin"),
        ),
    )

    define_attribute(
        _ATTRS,
        template=ParameterTemplate(
            name="Maximum Load Step",
            bounds=RealBounds(0, 200, "Pa"),
        ),
    )

    finalize_template(_ATTRS, TEMPLATE)



from typing import ClassVar

from gemd import (
    MaterialTemplate,
    PropertyTemplate,
    CategoricalBounds,
    NominalCategorical,
)
from gemd.entity.bounds import RealBounds
from gemd.entity.value import NominalReal

from openmsimodel.entity.base.material import Material
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
)

__all__ = ["Composition"]


class Composition(Material):
    """Class representing composition"""

    TEMPLATE: ClassVar[MaterialTemplate] = MaterialTemplate(
        "Alloy Composition",
        description="Composition selected to be fabricated and characterized",
    )

    _ATTRS: ClassVar[AttrsDict] = {"properties": {}}

    # define_attribute(
    #     _ATTRS,
    #     template=PropertyTemplate(
    #         'Al',
    #         bounds=RealBounds(0,100,''),
    #         description='Percentage of the total composition that Al takes'
    #     ),
    #     default_value=NominalReal(0.0,'')
    # )
    # define_attribute(
    #     _ATTRS,
    #     template=PropertyTemplate(
    #         'Cr',
    #         bounds=RealBounds(0,100,''),
    #         description='Percentage of the total composition that Cr takes'
    #     ),
    #     default_value=NominalReal(0.0,'')
    # )
    # define_attribute(
    #     _ATTRS,
    #     template=PropertyTemplate(
    #         'Co',
    #         bounds=RealBounds(0,100,''),
    #         description='Percentage of the total composition that Co takes'
    #     ),
    #     default_value=NominalReal(0.0,'')
    # )
    # define_attribute(
    #     _ATTRS,
    #     template=PropertyTemplate(
    #         'Fe',
    #         bounds=RealBounds(0,100,''),
    #         description='Percentage of the total composition that Fe takes'
    #     ),
    #     default_value=NominalReal(0.0,'')
    # )
    # define_attribute(
    #     _ATTRS,
    #     template=PropertyTemplate(
    #         'Ni',
    #         bounds=RealBounds(0,100,''),
    #         description='Percentage of the total composition that Ni takes'
    #     ),
    #     default_value=NominalReal(0.0,'')
    # )
    # define_attribute(
    #     _ATTRS,
    #     template=PropertyTemplate(
    #         'V',
    #         bounds=RealBounds(0,100,''),
    #         description='Percentage of the total composition that V takes'
    #     ),
    #     default_value=NominalReal(0.0,'')
    # )

    finalize_template(_ATTRS, TEMPLATE)



"""Base class for classes containing GEMD templates and objects."""

import os, warnings
from abc import ABC, abstractmethod
from typing import ClassVar, Type, Optional

from gemd import FileLink
from gemd.entity.attribute.base_attribute import BaseAttribute
from gemd.entity.util import make_instance

# from gemd.util.impl import set_uuids
from openmsimodel.entity.impl import assign_uuid

from openmsimodel.utilities.cached_isinstance_functions import (
    isinstance_template,
    isinstance_spec,
    isinstance_run,
)
from openmsimodel.utilities.typing import (
    Template,
    Spec,
    Run,
    SpecOrRun,
    SpecOrRunLiteral,
)
from openmsimodel.utilities.attributes import (
    AttrsDict,
    validate_state,
    update_attrs,
    remove_attrs,
    _validate_temp_keys,
    define_attribute,
    finalize_template,
)

from openmsimodel.utilities.logging import Logger
import openmsimodel.stores.gemd_template_store as store_tools

__all__ = ["BaseElement"]


class BaseElement(ABC):
    """
    Base class for materials, processes, and measurements.

    ``BaseElement`` 's are thin wrappers for GEMD entities. One ``BaseElement`` contains
    a template, spec, and run for the same kind of entity (``Material``,
    ``Process``, or ``Measurement``) and helps to create, update, link, and
    output these.

    Note that `name` is the GEMD name given to the spec and run. The template
    name is the name of the subclass.

    To subclass:

    1. Instantiate ``TEMPLATE`` as follows:
    ``TEMPLATE: ClassVar[Template] = Template(name=__name__)``,
    replacing ``Template`` with one of ``MaterialTemplate``,
    ``ProcessTemplate``, or ``MeasurementTemplate``.

    2. Instantiate ``_ATTRS`` as follows:
    ``_ATTRS``: ClassVar[AttrsDict] = _validate_temp_keys(TEMPLATE)

    3. Add conditions, parameters, and/or properties using
    ``define_attribute(_ATTRS, ...)`` from the ``qf_gemd.base.attributes``
    module.

    4. Call ``finalize_template(_ATTRS, TEMPLATE)``, found in the
    ``qf_gemd.base.attributes`` module, to add attributes to ``TEMPLATE``.

    5. Follow any additional subclass directions.
    """

    _TempType: ClassVar[Type[Template]]
    _SpecType: ClassVar[Type[Spec]]
    _RunType: ClassVar[Type[Run]]

    TEMPLATE: ClassVar[Template]

    _ATTRS: ClassVar[AttrsDict]

    # TODO: remove and put somewhere
    _TAG_SEP: ClassVar[str] = "::"

    def __init__(
        self,
        name: str,
        *,
        template: ClassVar[
            Template
        ] = None,  # TODO: triple check with abc's hastemplate
        notes: Optional[str] = None,
    ) -> None:
        super().__init__()

        self.name = name
        self.logger = Logger()
        self.TEMPLATE_WRAPPER = {}
        has_template = hasattr(self, "TEMPLATE")

        if not has_template and not template:
            # TODO: Change msg
            raise AttributeError(
                f"TEMPLATE is not defined. Assign to 'template' parameter an instance of either {Template.__dict__['__args__']},\n OR create a new subclass with a defined TEMPLATE attribute."
            )

        if template:
            if has_template:
                warnings.warn(
                    f"Found template '{self.TEMPLATE.name}', but '{template.name}' will be used instead.",
                    ResourceWarning,
                )
            self.TEMPLATE = template
            # if ("persistent_id" in self.TEMPLATE.uids.keys()) or (
            #     "auto" in self.TEMPLATE.uids.keys()
            # ):
            #     raise KeyError(
            #         f'the "auto" and "persistent_id" uid keys are reserved. Use another key. '
            #     )

            # TODO: Extend (or sync with external func that returns a dict for runs/specs

        # TODO: change from file when supporting file links
        # registering the object templates
        if store_tools.stores_config.activated:
            for i, store in enumerate(
                store_tools.stores_config.all_template_stores.values()
            ):
                self.TEMPLATE_WRAPPER[store.id] = store.register_new_template(
                    self.TEMPLATE,
                    from_file=False,
                    from_store=False,
                    from_memory=bool(template),
                    from_subclass=bool(has_template and not template),
                )
                if store_tools.stores_config.designated_store_id == store.id:
                    self.TEMPLATE = self.TEMPLATE_WRAPPER[
                        store_tools.stores_config.designated_store_id
                    ].template

        if (
            template
        ):  # FIXME: maybe dont need to check, just call all the time? even if redundant?
            self.prepare_attrs()

        # registering the attribute templates
        if store_tools.stores_config.activated:
            for i, store in enumerate(
                store_tools.stores_config.all_template_stores.values()
            ):
                for _attr_type in self._ATTRS.keys():
                    for _attr in self._ATTRS[_attr_type].values():
                        store_tools.stores_config.all_template_stores[
                            store.id
                        ].register_new_template(_attr["obj"])

        self._spec: Spec = self._SpecType(
            name=name, notes=notes, template=self.TEMPLATE
        )
        self._run: Run = make_instance(self._spec)
        assign_uuid(self._spec, "auto")
        assign_uuid(self._run, "auto")  # redundant?

    @property
    @abstractmethod
    def spec(self) -> Spec:
        """The underlying GEMD spec."""

    @property
    @abstractmethod
    def run(self) -> Run:
        """The underlying GEMD run."""

    ############################### ATTRIBUTES ###############################

    def prepare_attrs(self):
        self._ATTRS = _validate_temp_keys(self.TEMPLATE)
        if hasattr(self.TEMPLATE, "conditions"):
            for c in self.TEMPLATE.conditions:
                define_attribute(
                    self._ATTRS, template=c[0]
                )  # TODO: look into this weird format from GEMD (attr, bounds)
        if hasattr(self.TEMPLATE, "parameters"):
            for pa in self.TEMPLATE.parameters:
                define_attribute(self._ATTRS, template=pa[0])
        if hasattr(self.TEMPLATE, "properties"):
            for pr in self.TEMPLATE.properties:
                define_attribute(self._ATTRS, template=pr[0])
        # finalize_template(self._ATTRS, self.TEMPLATE)

    def _update_attributes(
        self,
        AttrType: Type[BaseAttribute],
        attributes: tuple[BaseAttribute],
        replace_all: bool = False,
        which: SpecOrRunLiteral = "spec",
    ) -> None:
        """Update attributes and link attribute templates."""

        update_attrs(
            self._ATTRS, self._spec, self._run, AttrType, attributes, replace_all, which
        )

    def _remove_attributes(
        self,
        AttrType: Type[BaseAttribute],
        attr_names: tuple[str, ...],
        which: SpecOrRunLiteral = "spec",
    ) -> None:
        """Remove attributes by name."""

        remove_attrs(self._ATTRS, self._spec, self._run, AttrType, attr_names, which)

    ############################### TAGS ###############################
    def update_tags(
        self,
        *tags: tuple[str, ...],
        replace_all: bool = False,
        which: SpecOrRunLiteral = "spec",
    ) -> None:
        """
        Change or add hierarchical tags.

        Each tag is represented by a ``tuple`` of hierarchical ``str`` s. For
        example, ``('Quantum Design', 'MPMS3')``, in that order, describes the
        make and model of a particular measurement instrument.

        When `replace_all` is ``False``, any existing tags that are exactly the
        same as a new tag will not be duplicated. However, a tag that is the
        "child" of an existing tag will not override the "parent" tag. For
        example, it is possible to have all of the following at once:
        ``('Quantum Design', 'DynaCool')``,
        ``('Quantum Design', 'DynaCool', '1')``,
        and ``('Quantum Design', 'DynaCool', '3')``.

        Internally, tag strings will be concatenated with ``'::'`` as
        recommended by the GEMD specification.

        tags: tuple[str]
            tuples representing tags to add. Each tuple should contain the
            components of a tag from most general to most specific.
        replace_all: bool, default False
            If ``True``, remove any existing tags before adding new ones.
        which: {'spec', 'run', 'both'}, default 'spec'
            Whether to update the spec, run, or both.
        """

        validate_state(which)

        if which in ["spec", "both"]:
            self._set_tags(self._spec, tags, replace_all)

        if which in ["run", "both"]:
            self._set_tags(self._run, tags, replace_all)

    def remove_tags(
        self, *tags: tuple[str, ...], which: SpecOrRunLiteral = "spec"
    ) -> None:
        """Remove tags.

        See `update_tags` for tag format details. Tags are removed by exact
        comparison of the underlying hierarchcal ``str`` s.

        tags: tuple[str]
        ``tuple`` s representing tags to remove.
        which: {'spec', 'run', 'both'}, default 'spec'
        Whether to remove from the spec, run, or both.
        """

        validate_state(which)

        if which in ["spec", "both"]:
            self._remove_tags(self._spec, tags)

        if which in ["run", "both"]:
            self._remove_tags(self._run, tags)

    # TODO: merge into one method?
    @classmethod
    def _set_tags(
        cls,
        spec_or_run: SpecOrRun,
        tags: tuple[tuple[str, ...], ...],
        replace_all: bool = False,
    ) -> None:
        """Set tags for the spec or run."""

        tag_strs = [cls._TAG_SEP.join(tag) for tag in tags]

        if not replace_all:
            existing_tags = [tag for tag in spec_or_run.tags if tag not in tag_strs]
        else:
            existing_tags = []

        spec_or_run.tags = existing_tags + tag_strs

    @classmethod
    def _remove_tags(
        cls, spec_or_run: SpecOrRun, tags: tuple[tuple[str, ...], ...]
    ) -> None:
        """Remove tags from the spec or run."""

        tag_strs = [cls._TAG_SEP.join(tag) for tag in tags]
        spec_or_run.tags = [tag for tag in spec_or_run.tags if tag not in tag_strs]

    @staticmethod
    def _build_tags_dict(tags: list[str], parent_dict: dict, tag_sep: str) -> None:
        """Build a spec or run hierarchical tags ``dict``."""

        for tag_str in tags:
            tag_tup = tag_str.split(tag_sep)
            parent = parent_dict
            for component in tag_tup:
                if component not in parent:
                    parent[component] = {}
                parent = parent[component]

    def get_tags_dict(self):
        """Get a ``dict`` representing the hierarchical tags."""

        tags_dict = {"spec": {}, "run": {}}

        self._build_tags_dict(self._spec.tags, tags_dict["spec"], self._TAG_SEP)
        self._build_tags_dict(self._run.tags, tags_dict["run"], self._TAG_SEP)

        return tags_dict

    ############################### FILE LINKS ###############################
    def update_filelinks(
        self,
        *filelinks: FileLink,
        replace_all: bool = False,
        which: SpecOrRunLiteral = "spec",
    ) -> None:
        """
        Change or add file links.

        filelinks: FileLink
        The file links to change or add.
        replace_all: bool, default False
        If ``True``, remove any existing file links before adding new ones.
        which: {'spec', 'run', 'both'}, default 'spec'
        Whether to update the spec, run, or both.
        """

        validate_state(which)

        supplied_links = {self._link_str(link): link for link in filelinks}

        if which in ["spec", "both"]:
            self._set_filelinks(self._spec, supplied_links, replace_all)

        if which in ["run", "both"]:
            self._set_filelinks(self._run, supplied_links, replace_all)

    @classmethod
    def _set_filelinks(
        cls,
        spec_or_run: SpecOrRun,
        supplied_links: dict[str, FileLink],
        replace_all: bool = False,
    ) -> None:
        """Set links for the spec or run."""

        existing_links = (
            {}
            if replace_all
            else {cls._link_str(link): link for link in spec_or_run.file_links}
        )

        spec_or_run.file_links = {**existing_links, **supplied_links}.values()

    def remove_filelinks(
        self, *filelinks: FileLink, which: SpecOrRunLiteral = "spec"
    ) -> None:
        """Remove file links.

        filelinks: tuple[str]
            The file links to remove by comparison of the underlying url and
            filename.
        which: {'spec', 'run', 'both'}, default 'spec'
            Whether to remove from the spec, run, or both.
        """

        validate_state(which)

        if which in ["spec", "both"]:
            self._remove_filelinks(self._spec, filelinks)

        if which in ["run", "both"]:
            self._remove_filelinks(self._run, filelinks)

    @classmethod
    def _remove_filelinks(
        cls, spec_or_run: SpecOrRun, filelinks: tuple[FileLink]
    ) -> None:
        """Remove tags from the spec or run."""

        filelink_strs = [cls._link_str(link) for link in filelinks]

        spec_or_run.file_links = [
            link
            for link in spec_or_run.file_links
            if cls._link_str(link) not in filelink_strs
        ]

    @staticmethod
    def _link_str(link: FileLink) -> str:
        """
        Return a str representation of a ``FileLink`` based on its ``filename`` and url``.
        """
        # TODO: add conversion to standard file path or https
        return f'{link.filename}{"/" if link.filename.endswith("/") else ","}{link.url}'

    def get_filelinks_dict(self):
        """
        Get string representations of the file links.

        filelinks_dict: FileLinksDict
            Strings representing the file links of the spec and run.
        """

        filelinks_dict = {}

        filelinks_dict["spec"] = tuple(
            self._link_str(link) for link in self._spec.file_links
        )
        filelinks_dict["run"] = tuple(
            self._link_str(link) for link in self._run.file_links
        )

        return filelinks_dict

    # TODO: add a 'update_spec' function to != the 1:1 / template:spec duality
    def return_all_gemd(self) -> list:
        if self.__class__.__name__ == "Ingredient":
            return [self.spec, self.run]
        else:
            return [self.TEMPLATE, self.spec, self.run]

    # @classmethod
    # def from_(cls, any):
    #     if isinstance_run(any):
    #         # base_element = cls(any.name, template=any.template)
    #         # base_element.spec = any.spec
    #         # base_element.run = any
    #     elif isinstance_spec(any):
    #         base_element = cls(any.name, template=any.template)
    #         base_element.spec = any
    #     elif isinstance_template(any):
    #         base_element = cls(any.name, template=any)
    #     return base_element

    # @abstractmethod
    # def to_form(self) -> str:
    #     """Return a ``str`` specifying how to create a web form for this node."""

    # def thin_dumps(self, encoder, destination):
    #     for obj in [self._spec, self._run]:
    #         encoder.thin_dumps(obj,indent=3) # trigger uids assignment
    #         fn = '_'.join([obj.__class__.__name__, obj.name,obj.uids['auto']])
    #         with open(os.path.join(destination, fn),'w') as fp:
    #             fp.write(encoder.thin_dumps(obj,indent=3))

    # def thin_dumps(self, encoder, destination):
    #     self.dump_loop(encoder.thin_dumps, destination)

    # def dumps(self, encoder, destination):
    #     self.dump_loop(encoder.dumps, destination)

    # def dump_loop(self, encoder_func, destination):
    #     for obj in [self._spec, self._run]:
    #         encoder_func(obj, indent=3)  # trigger uids assignment
    #         fn = "_".join([obj.__class__.__name__, obj.name, obj.uids["auto"]])
    #         with open(os.path.join(destination, fn), "w") as fp:
    #             fp.write(encoder_func(obj, indent=3))



from openmsimodel.entity.base.material import Material
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
    _validate_temp_keys,
)

from typing import ClassVar

from gemd import (
    MaterialTemplate,
    PropertyTemplate,
    CategoricalBounds,
    NominalCategorical,
    RealBounds,
    NominalReal,
)


class Alloy(Material):
    """Class representing an Alloy"""

    TEMPLATE: ClassVar[MaterialTemplate] = MaterialTemplate(
        name="Alloy",
        description="Alloy generated with a unique composition and synthesis parameters",
    )

    _ATTRS: ClassVar[AttrsDict] = _validate_temp_keys(TEMPLATE)

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Form", bounds=CategoricalBounds(categories=["Ingot"])
        ),
    )

    finalize_template(_ATTRS, TEMPLATE)



from typing import ClassVar

from gemd import (
    ProcessTemplate,
    ParameterTemplate,
    CategoricalBounds,
    NominalCategorical,
)

from openmsimodel.entity.base.process import Process
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
    _validate_temp_keys,
)


class AddMaterial(Process):
    """Class representing the aggregation of materials/elements for an experiment"""

    TEMPLATE: ClassVar[ProcessTemplate] = ProcessTemplate(
        name="Add Material",
        description="""Adding the correct ingredients to produce a mixture that can go through
                fabrication procedure
                """,
    )

    _ATTRS: ClassVar[AttrsDict] = _validate_temp_keys(TEMPLATE)
    # define_attribute(
    #     _ATTRS,
    #     template=ParameterTemplate(
    #         name='Supplier',
    #         bounds=CategoricalBounds(categories=[''])
    #     ),
    #     default_value=NominalCategorical('')
    # )

    finalize_template(_ATTRS, TEMPLATE)



from openmsimodel.entity.base.material import Material
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
    _validate_temp_keys
)

from typing import ClassVar

from gemd import (
    MaterialTemplate,
    PropertyTemplate,
    CategoricalBounds,
    NominalCategorical,
    RealBounds,
    NominalReal,
)

class TravelerSamples(Material):
    """Class representing grouped traveler samples"""

    TEMPLATE: ClassVar[MaterialTemplate] = MaterialTemplate(
        name="Traveler Samples",
    )

    _ATTRS: ClassVar[AttrsDict] = _validate_temp_keys(TEMPLATE)

    finalize_template(_ATTRS, TEMPLATE)



from pathlib import Path
from collections import defaultdict
import os
import shutil
import argparse
import json
from .folder_or_file import FolderOrFile

# import sys
# sys.path.append("..")
# from openmsimodel.utilities.tools import *
from openmsimodel.utilities.runnable import Runnable
from openmsimodel.utilities.argument_parsing import OpenMSIModelParser

from gemd.json import GEMDJson
from gemd.util.impl import recursive_foreach


# TODO: extend Logging
class Workflow(Runnable):
    """Class to model a workflow, typically a set of processing steps, experiments, and characterizations, into GEMD, a data model.
    the definition of a workflow is meant to be flexible to the needs of the user. Workflows can be composed
    to construct even larger GEMD graphs.
    It offers utilities functions to build the model flexibly, break it down into smaller, easier to
    manage blocks, or complete operations such as dumping and loading models into/from JSONs.
    """

    ARGUMENT_PARSER_TYPE = OpenMSIModelParser

    def __init__(self, *args, **kwargs):
        """Initialization of workflow"""
        # self.root
        self.elements = []
        self.subs = defaultdict()
        self.terminal_subs = defaultdict()
        # self.blocks = defaultdict()
        # self.terminal_blocks = defaultdict()
        # self.encoder = GEMDJson()
        # if "output" in args:
        #     self.output = args.output

    def build(self):
        """
        This function builds the entire GEMD model that represents a certain Workflow.
        to be overwritten by child objects of Workflow that correspond to a specific workflow / user case.
        """
        pass

    def thin_dumps(self):
        """
        dumps the entire model into a JSON per object, each representing the 'thin' version' of the object
        in which pointers (i.e., true value) are replaced by links (e.g., uuid).
        """
        pass

    def dumps(self):
        """
        dumps the entire model into a single JSON, which contains all the model objects with data pointers (!= links).
        """
        pass

    def thin_loads(self):
        """
        loads the entire model from a list of JSONs, each representing the 'thin' version' of the model object
        in which pointers (i.e., true value) are replaced by links (e.g., uuid).
        """
        pass

    def loads(self):
        """
        loads the entire model from a single JSON, which contains all the model objects with data pointers (!= links)
        """
        pass

    def local_out(self, item):
        """
        function object to run on individual item during recursion
        :param item: json item to write its destination
        se
        """
        fn = "_".join([item.__class__.__name__, item.name, item.uids["auto"], ".json"])
        with open(os.path.join(self.local_out_destination, fn), "w") as fp:
            # if self.dump_function.__name__ == "dumps":
            #     result = json.loads(self.dump_function(item, indent=3))
            #     fp.write(result)
            # else:
            fp.write(self.dump_function(item, indent=3))

    #################### CLASS METHODS ####################

    @classmethod
    def get_command_line_arguments(cls):
        superargs, superkwargs = super().get_command_line_arguments()
        args = []
        kwargs = {**superkwargs}
        return args, kwargs

    @classmethod
    def run_from_command_line(cls, args=None):
        parser = cls.get_argument_parser()
        args = parser.parse_args(args=args)
        workflow = cls(*args)
        workflow.build()


def main(args=None):
    """
    Main method to run from command line
    """
    Workflow.run_from_command_line(args)


if __name__ == "__main__":
    main()



"""Attribute functions and templates used elsewhere."""

from typing import TypedDict, Type, Union, Optional, Any, get_args

from gemd import (
    ConditionTemplate,
    ParameterTemplate,
    PropertyTemplate,
    Condition,
    Parameter,
    Property,
    PropertyAndConditions,
    ProcessTemplate,
    MaterialTemplate,
    MeasurementTemplate,
)
from gemd.entity.template.attribute_template import AttributeTemplate
from gemd.entity.attribute.base_attribute import BaseAttribute
from gemd.entity.bounds.base_bounds import BaseBounds
from gemd.entity.value.base_value import BaseValue
from gemd.enumeration import Origin

from .typing import Template, Spec, Run, Attributes, SpecOrRun, SpecOrRunLiteral


class AttrDict(TypedDict):
    """
    Define dictionary for specifying required/optional attribute templates with defaults.

    If ``bounds`` is not ``None``, it will be added as a more restrictive set
    of bounds to the attribute in the template.

    If ``default_value`` is ``None``, the attribute is optional. Otherwise, it
    is required and should be added to the spec upon instantiation with the
    default value (in subclasses).
    """

    bounds: Optional[BaseBounds]
    default_value: Optional[BaseValue]


class CondDict(AttrDict):
    """Add condition template key."""

    cond: ConditionTemplate


class ParamDict(AttrDict):
    """Add parameter template key."""

    param: ParameterTemplate


class PropDict(AttrDict):
    """Add property template key."""

    prop: PropertyTemplate


class AttrsDict(TypedDict, total=False):
    """Define dictionary containing local attributes."""

    conditions: dict[str, CondDict]
    parameters: dict[str, ParamDict]
    properties: dict[str, PropDict]


def define_attribute(
    cls_attrs: AttrsDict,
    template: AttributeTemplate,
    bounds: Optional[BaseBounds] = None,
    default_value: Optional[BaseValue] = None,
) -> None:
    """
    Add attribute `template` to `cls_attrs`.

    Use as needed at the beginning of each subclass.
    """

    if isinstance(template, ConditionTemplate):
        key1 = "conditions"
        # key2 = "cond"
    elif isinstance(template, ParameterTemplate):
        key1 = "parameters"
        # key2 = "param"
    elif isinstance(template, PropertyTemplate):
        key1 = "properties"
        # key2 = "prop"
    else:
        raise TypeError(
            "template must be an instance of ConditionTemplate, "
            "ParameterTemplate, or PropertyTemplate."
        )
    key2 = "obj"

    if bounds is not None and not isinstance(bounds, BaseBounds):
        raise TypeError("bounds must be an instance of BaseBounds.")

    if default_value is not None and not isinstance(default_value, BaseValue):
        raise TypeError("default_value must be an instance of BaseValue.")

    if not (key1 in cls_attrs.keys()):
        raise TypeError(
            f'attribute template of type "{key1}" cannot be assigned to attributes dictionary with keys "{cls_attrs.keys()}".'
        )

    cls_attrs[key1][template.name] = {
        key2: template,
        "bounds": bounds,
        "default_value": default_value,
    }


def finalize_template(cls_attrs: AttrsDict, cls_template: Template) -> None:
    """
    Add conditions, parameters, and properties to ``cls_template``.

    Use in subclasses after all ``define_attribute`` calls.
    """

    if cls_attrs.get("conditions") is not None:
        cls_template.conditions = [
            (cond_dict["obj"], cond_dict["bounds"])
            for cond_dict in cls_attrs["conditions"].values()
        ]

    if cls_attrs.get("parameters") is not None:
        cls_template.parameters = [
            (param_dict["obj"], param_dict["bounds"])
            for param_dict in cls_attrs["parameters"].values()
        ]

    if cls_attrs.get("properties") is not None:
        cls_template.properties = [
            (prop_dict["obj"], prop_dict["bounds"])
            for prop_dict in cls_attrs["properties"].values()
        ]


def update_attrs(
    attrs: AttrsDict,
    spec: Spec,
    run: Run,
    AttrType: Type[Union[BaseAttribute, PropertyAndConditions]],
    attributes: tuple[BaseAttribute],
    replace_all: bool = False,
    which: SpecOrRunLiteral = "spec",
) -> None:
    """Used by BaseElement to update attributes and link attribute templates."""

    attr_dict_key, singular, plural = _validate_attr_type(AttrType)
    validate_state(which)

    required_attrs = _required_attrs(attrs, AttrType, attr_dict_key, plural)

    # for attr in attributes:
    #     if attr.name not in attrs[plural]:
    #         raise ValueError(f'{singular.capitalize()} "{attr.name}" is not supported.')

    supplied_attrs = {attr.name: attr for attr in attributes}

    for attr_name, attr in supplied_attrs.items():
        # TODO: quick fix for now for having multiple version of a same template (i.e., ambient ressure, purging pressure for the same parameter template: pressure )
        # TODO: applies to uncommented bloc of code above that caused errors
        if attr_name not in attrs[plural].keys():  # TODO: improve fix?
            if (
                hasattr(attr, "template")
                and attr.template.name not in attrs[plural].keys()
            ):
                raise KeyError(
                    f"the '{attr_name}' attribute is not among the object defined attributes."
                )
            attr_name = attr.template.name  # FIXME
        # if attr.template is not None or (attr.property is not None and attr.property.template is not None): #FIXME
        #     continue
        if type(attr) == PropertyAndConditions:
            if hasattr(attr, "property") and attr.property.template is not None:
                attr.property.template = attrs[plural][attr_name]["obj"]
        else:
            if attr.template is not None:
                attr.template = attrs[plural][attr_name][
                    "obj"
                ]  # TODO: fix attr_name, change to temp_name and set temp_name to whole name or to template.name based on if found

    if which in ["spec", "both"]:
        _set_attrs(spec, required_attrs, supplied_attrs, AttrType, replace_all)
    if which in ["run", "both"]:
        _set_attrs(run, required_attrs, supplied_attrs, AttrType, replace_all)


def _required_attrs(
    attrs: AttrsDict,
    AttrType: Type[Union[BaseAttribute, PropertyAndConditions]],
    attr_dict_key: str,
    plural: str,
) -> Union[BaseAttribute, PropertyAndConditions]:
    """
    Return dict of required attributes, handling case of ``PropertyAndConditions`` for materials.
    """

    if AttrType is PropertyAndConditions:
        return {
            attr_name: PropertyAndConditions(
                property=Property(
                    name=attr_name,
                    value=attr_dict["default_value"],
                    template=attr_dict["obj"],
                    origin=Origin.SPECIFIED,
                )
            )
            for attr_name, attr_dict in attrs[plural].items()
            if attr_dict["default_value"] is not None
        }
    return {
        attr_name: AttrType(
            name=attr_name,
            value=attr_dict["default_value"],
            template=attr_dict["obj"],
            origin=Origin.SPECIFIED,
        )
        for attr_name, attr_dict in attrs[plural].items()
        if attr_dict["default_value"] is not None
    }


def _set_attrs(
    spec_or_run: SpecOrRun,
    required_attrs: dict[str, BaseAttribute],
    supplied_attrs: dict[str, BaseAttribute],
    AttrType: Type[Union[BaseAttribute, PropertyAndConditions]],
    replace_all: bool,
) -> None:
    """Set spec or run attributes appropriately, depending on `AttrType`."""

    if AttrType is Condition:
        existing_attrs = (
            {} if replace_all else {attr.name: attr for attr in spec_or_run.conditions}
        )
        spec_or_run.conditions = {
            **required_attrs,
            **existing_attrs,
            **supplied_attrs,
        }.values()
    elif AttrType is Parameter:
        existing_attrs = (
            {} if replace_all else {attr.name: attr for attr in spec_or_run.parameters}
        )
        spec_or_run.parameters = {
            **required_attrs,
            **existing_attrs,
            **supplied_attrs,
        }.values()
    elif AttrType in [Property, PropertyAndConditions]:
        existing_attrs = (
            {} if replace_all else {attr.name: attr for attr in spec_or_run.properties}
        )
        spec_or_run.properties = {
            **required_attrs,
            **existing_attrs,
            **supplied_attrs,
        }.values()


def remove_attrs(
    attrs: AttrsDict,
    spec: Spec,
    run: Run,
    AttrType: Type[Union[BaseAttribute, PropertyAndConditions]],
    attr_names: tuple[str, ...],
    which: SpecOrRunLiteral = "spec",
) -> None:
    """Used by BaseElement to remove attributes by name."""

    _, singular, plural = _validate_attr_type(AttrType)
    validate_state(which)

    required_names = [
        attr_name
        for attr_name, attr_dict, in attrs[plural].items()
        if attr_dict["default_value"] is not None
    ]

    for name in attr_names:
        if name not in attrs[plural]:
            raise ValueError(f'{singular.capitalize()} "{name}" is not supported.')
        if name in required_names:
            raise ValueError(f'May not remove required {singular} "{name}".')

    if which in ["spec", "both"]:
        _remove_attrs(AttrType, spec, attr_names)
    if which in ["run", "both"]:
        _remove_attrs(AttrType, run, attr_names)


def _remove_attrs(
    AttrType: Type[
        Union[BaseAttribute, PropertyAndConditions]
    ],  # TODO: figure whether to do this or use 'Attributes' from typing
    spec_or_run: SpecOrRun,
    attr_names: tuple[str, ...],
) -> None:
    """Remove spec or run attributes appropriately, depending on `AttrType`."""

    if AttrType is Condition:
        left_over = [
            attr for attr in spec_or_run.conditions if attr.name not in attr_names
        ]
        spec_or_run.conditions = left_over
    elif AttrType is Parameter:
        left_over = [
            attr for attr in spec_or_run.parameters if attr.name not in attr_names
        ]
        spec_or_run.parameters = left_over
    elif AttrType in [Property, PropertyAndConditions]:
        left_over = [
            attr for attr in spec_or_run.properties if attr.name not in attr_names
        ]
        spec_or_run.properties = left_over


def _validate_temp_keys(Template: Template) -> AttrsDict:
    """Check `Template`, or Template type, and returns the corresponding AttrsDict."""
    # ATTRS = {}
    if isinstance(Template, ProcessTemplate):
        ATTRS = {"conditions": {}, "parameters": {}}
    elif isinstance(Template, MaterialTemplate):
        ATTRS = {"properties": {}}
    elif isinstance(Template, MeasurementTemplate):
        ATTRS = {"properties": {}, "conditions": {}, "parameters": {}}
    else:
        raise ValueError(
            "AttrType must be one of ProcessTemplate, MaterialTemplate, or MeasurementTemplate."
        )
    return AttrsDict(ATTRS)


def _validate_attr_type(AttrType: Attributes) -> None:
    """Check `AttrType`, return case-specific strings."""

    if AttrType is Condition:
        attr_dict_key = "cond"
        singular = "condition"
        plural = "conditions"
    elif AttrType is Parameter:
        attr_dict_key = "param"
        singular = "parameter"
        plural = "parameters"
    elif AttrType in [Property, PropertyAndConditions]:
        attr_dict_key = "prop"
        singular = "property"
        plural = "properties"
    else:
        raise ValueError(
            "AttrType must be one of Condition, Parameter, Property, or PropertyAndConditions."
        )

    return attr_dict_key, singular, plural


def validate_state(which: Any) -> None:
    """Validate `which`."""

    if which not in get_args(SpecOrRunLiteral):
        raise ValueError(f"{which} must be one of {get_args(SpecOrRunLiteral)}")



from sqlalchemy.orm import Session
from sqlalchemy.ext.automap import automap_base
from openmsimodel.utilities.tools import read_gemd_data
from gemd.json import GEMDJson
import json

model_table_name = "GEMDModel"
object_table_name = "GEMDObject"
attribute_table_name = "GEMDAttribute"
edge_table_name = "GEMDEdge"


def show_models():
    """shows all models in the database.

    Returns:
        _type_: str
    """
    return """select distinct * from GEMDModel"""


def top_elements(model_id, nb, gemd_type):
    """return top elements of a certain type from model.

    Args:
        model_id (int): id of the model to query
        nb (int): number of elements
        gemd_type (str): type of gemd object

    Returns:
        _type_: str
    """
    return f"""select top {nb} context from gemdobject c where gemd_type='{gemd_type}' AND c.model_id='{model_id}' order by newid()"""


def display_all(model_id, type_to_display):
    """displays all the gemd objects of a certain type in the model.

    Args:
        model_id (int): id of the model to query
        type_to_display (str): object type to display

    Returns:
        _type_: str
    """
    return f""" select * from {type_to_display} where model_id={model_id}"""


def return_all_paths(model_id):
    """return all paths between all nodes in model"""
    return f"""
    with gr as (
    select c.uid as root_uid
    ,      c.gemd_type as root_type
    ,      0 as level
    ,      cast(NULL as varchar(64)) as endpoint_uid
    ,      c.uid as from_uid, cast(NULL as bigint) as edge_id, cast(NULL as varchar(64)) as gemd_ref
    ,      cast(gemd_type+':'+c.uid as varchar(max)) as [path]
    from GEMDObject c where c.model_id={model_id} 
    union all
    select gr.root_uid, gr.root_type, gr.level+1, e.to_uid
    ,      e.to_uid, e.id, e.gemd_ref
    ,      gr.path+'==>'+e.gemd_ref+':'+e.to_uid
    from gr
    join GEMDEdge e on e.from_uid=gr.from_uid
    where gr.level < 16
    )
    select root_uid, root_type, endpoint_uid
    ,      edge_id,gemd_ref
    ,      path, level
    from gr
    order by root_type,root_uid, path
    """


def gemd_types_query():
    return """
    select distinct gemd_type
    ,      x.[key]
    from GEMDObject cross apply openjson(context,'$') x
    where x.[key] not in ('tags','description','uids','name','type','bounds','file_links','labels')
    order by 1,2
    """


def reachable_nodes_query(uid):
    """returns all the elements that can be reached from the material.

    Args:
        uid (str): unique identifier of object to query with

    Returns:
        _type_: str
    """
    return f"""
        with gr as (
        select c.uid as node_uid
        ,      c.gemd_type as node_type
        ,      c.context as node_context
        ,      cast(c.gemd_type+' ['+c.uid+']' as varchar(128)) as Target
        ,      cast(-1 as bigint) as edge_id
        ,      cast(NULL as varchar(64)) as from_uid
        ,      cast(NULL as varchar(128)) as Source
        ,      0 as level
        from GEMDObject c
        where uid='{uid}'
        union all
        select c.uid as node_uid
        ,      c.gemd_type as node_type
        ,      c.context as node_context
        ,      cast(c.gemd_type+' ['+c.uid+']' as varchar(128)) as Target
        ,      e.id as edge_id
        ,      gr.node_uid as from_uid
        ,      gr.Target as Source
        ,      gr.level+1 as level
        from gr
        join GEMDEdge e on e.from_uid=gr.node_uid
        join GEMDObject c on c.uid=e.to_uid
        where gr.level < 16
        )
        select Source,Target
        --,node_type,node_context
        from gr
        where Source is not null
    """


def to_node_query(model_id):
    """returns all elements that can reach a given node, for all nodes in the model.

    Args:
        model_id (str): id of the model to query from

    Returns:
        _type_: str
    """
    return f"""
        with gr as (
        select c.uid as root_uid
        ,      c.gemd_type as root_type
        ,      0 as level
        ,      cast(NULL as varchar(64)) as endpoint_uid
        ,      c.uid as from_uid, cast(NULL as bigint) as edge_id, cast(NULL as varchar(64)) as gemd_ref
        ,      cast(gemd_type+c.uid as varchar(max)) as [path]
        from GEMDObject c where c.model_id={model_id}
        union all
        select gr.root_uid, gr.root_type, gr.level+1, e.to_uid
        ,      e.to_uid, e.id, e.gemd_ref
        ,      gr.path+'==>'+e.gemd_ref+':'+e.to_uid
        from gr
        join GEMDEdge e on e.from_uid=gr.from_uid
        where gr.level < 16
        )
        select endpoint_uid, count(distinct root_uid) as num_in_nodes
        from gr
        group by endpoint_uid
        order by num_in_nodes desc
    """


def multiple_paths_nodes_query(model_id):
    return f"""
    with gr as (
    select c.uid as root_uid
    ,      c.gemd_type as root_type
    ,      0 as level
    ,      cast(NULL as varchar(64)) as endpoint_uid
    ,      cast(NULL as varchar(32)) as endpoint_type
    ,      c.uid as from_uid, cast(NULL as bigint) as edge_id, cast(NULL as varchar(64)) as gemd_ref
    ,      cast(gemd_type+c.uid as varchar(max)) as [path]
    from GEMDObject c where c.model_id={model_id} AND gemd_type='material_run' 
    union all
    select gr.root_uid, gr.root_type, gr.level+1, e.to_uid, c.gemd_type
    ,      e.to_uid, e.id, e.gemd_ref
    ,      gr.path+'==>'+e.gemd_ref+':'+e.to_uid
    from gr 
    join GEMDEdge e on e.from_uid=gr.from_uid
    join GEMDObject c on c.uid=e.to_uid
    where gr.level < 16
    )
    select root_uid, root_type, endpoint_uid, endpoint_type
    ,      min(path) as path, min(level) as min_level, max(level) as max_level
    ,      count(*) as num_paths
    from gr
    group by root_type, root_uid, endpoint_uid,endpoint_type having count(*) > 1  -- if you want to find multiple paths between nodes
    order by root_type,root_uid, endpoint_uid,path
    """


def return_all_paths(model_id):
    """return all paths between all nodes, if exst.

    Args:
        model_id (str): id of the model to query from

    Returns:
        _type_: str
    """
    return f"""
        with gr as (
        select c.uid as root_uid
        ,      c.gemd_type as root_type
        ,      0 as level
        ,      cast(NULL as varchar(64)) as endpoint_uid
        ,      c.uid as from_uid, cast(NULL as bigint) as edge_id, cast(NULL as varchar(64)) as gemd_ref
        ,      cast(gemd_type+':'+c.uid as varchar(max)) as [path]
        from GEMDObject c where c.model_id={model_id} 
        union all
        select gr.root_uid, gr.root_type, gr.level+1, e.to_uid
        ,      e.to_uid, e.id, e.gemd_ref
        ,      gr.path+'==>'+e.gemd_ref+':'+e.to_uid
        from gr
        join GEMDEdge e on e.from_uid=gr.from_uid
        where gr.level < 16
        )
        select root_uid, root_type, endpoint_uid
        ,      edge_id,gemd_ref
        ,      path, level
        from gr
        order by root_type,root_uid, path
        """


###############################


def create_tables_query():
    """creates a table with the designed GEMD schema."""

    return f"""
    create table {model_table_name} (
        id bigint identity(1,1) not null
    , name varchar(1024) not null
    , constraint  primary key(id)
    )

    create table {object_table_name} (
    uid varchar(64) not null
    , model_id bigint not null
    , gemd_type varchar(32) not null
    , context varchar(max) not null
    , constraint pk_GEMDContext primary key(uid)
    , constraint fk_gemdcontext_model foreign key (model_id) REFERENCES gemdmodel_temp(id)
    )
    
    CREATE TABLE [dbo].[{attribute_table_name}](
    [gemdobject_uid] [varchar](64) NOT NULL,
    [template_id] [nvarchar](128) NULL,
    [name] [nvarchar](128) NULL,
    [value] [nvarchar](4000) NULL,
    [value_type] [nvarchar](32) NULL,
    [value_units] [nvarchar](128) NULL,
    [attribute_type] [nvarchar](128) NULL,
    [property] [nvarchar](max) NULL,
    [conditions] [nvarchar](max) NULL
    )

    CREATE TABLE {edge_table_name}(
        id bigint IDENTITY(1,1)  not null
    , from_uid varchar(64) not null
    , to_uid varchar(64) not null
    , gemd_ref varchar(64) not null
    )
    """


def create_edges_query(model_id):
    """creates and stores 'GEMDEdge's based on pre-existing 'GEMDObject's"""
    return f"""insert into GEMDEdge
    select uid,material_run_uid, 'ingredient_run:material_run'
    from ingredient_run WHERE model_id={model_id} AND material_run_uid is not null
    union
    select uid,process_run_uid, 'ingredient_run:process_run'
    from ingredient_run WHERE model_id={model_id} AND process_run_uid is not null
    union
    select uid,spec_uid, 'ingredient_run:spec'
    from ingredient_run WHERE model_id={model_id} AND process_run_uid is not null
    union
    select uid,process_spec_uid, 'ingredient_spec:process_spec'
    from ingredient_spec WHERE model_id={model_id} AND process_spec_uid is not null
    union
    select uid,material_spec_uid, 'ingredient_spec:material_spec'
    from ingredient_spec WHERE model_id={model_id} AND material_spec_uid is not null
    union
    select uid,spec_uid, 'material_run:spec'
    from material_run WHERE model_id={model_id} AND spec_uid is not null
    union
    select uid, process_run_uid, 'material_run:process_run'
    from material_run WHERE model_id={model_id} AND process_run_uid is not null
    union
    select uid,process_spec_uid, 'material_spec:process_spec'
    from material_spec WHERE model_id={model_id} AND process_spec_uid is not null
    union
    select uid,template_uid, 'material_spec:template'
    from material_spec WHERE model_id={model_id} AND template_uid is not null
    union
    select uid,spec_uid, 'measurement_run:spec'
    from measurement_run WHERE model_id={model_id} AND spec_uid is not null
    union
    select uid,material_uid, 'measurement_run:material'
    from measurement_run WHERE model_id={model_id} AND material_uid is not null
    union
    select uid,template_uid, 'measurement_spec:template'
    from measurement_spec WHERE model_id={model_id} AND template_uid is not null
    union
    select uid,spec_uid, 'process_run:spec'
    from process_run WHERE model_id={model_id} AND spec_uid is not null
    union
    select uid,template_uid, 'process_spec:template'
    from process_spec WHERE model_id={model_id} AND template_uid is not null
    """


def create_attributes_query(model_id):
    return f"""
        with a as (
        select s.uid, s.gemd_type
        ,      sp.*
        from GEMDObject s where model_id={model_id}
        cross apply OPENJSON(s.context,'$.parameters') with (
            template_id nvarchar(128) '$.template.id',
            name nvarchar(128) '$.name',
            value_type nvarchar(max) '$.value.type' ,
            value_units nvarchar(max) '$.value.units' ,
            value nvarchar(max) '$.value' as json,
            attribute_type nvarchar(128) '$.type',
            parameter nvarchar(max) '$' as json) sp
        )
        insert into GEMDAttribute
        select a.uid as gemdcontext_uid, a.template_id, a.name
        ,      case when a.value_type = 'nominal_categorical' then json_value(a.value,'$.category')
                    when a.value_type = 'nominal_real' then json_value(a.value,'$.nominal')
                    when a.value_type = 'nominal_integer' then json_value(a.value,'$.nominal')
                    when a.value_type = 'nominal_composition' then json_value(a.value,'$')
            else NULL end as value
            , a.value_type
            , a.value_units
        ,      a.attribute_type
        ,      a.parameter, NULL
        from a 
    """


def prepare_classes(db_engine):
    try:
        Base = automap_base()
        Base.prepare(db_engine, reflect=True)
        for c in Base.classes:
            print(c)
        GEMDObject = Base.classes.GEMDObject
        GEMDModel = Base.classes.GEMDModel
        return GEMDObject, GEMDModel
    except Exception as e:
        raise e


def load_model_query(name, db, dirpath, uuid):
    """function to load a model into the base

    Args:
        name (str): name of the model
        dirpath (str): path to folder or single file containing JSONs
        uuid (str, optional): _description_.
    """
    print("Loading model and GEMDObjects...")

    GEMDObject, GEMDModel = prepare_classes(db.ENGINE)

    # creating GEMDModels
    model = GEMDModel(name=name)

    # creating GEMDObjects
    files = read_gemd_data(dirpath, GEMDJson())
    for f in files:
        try:
            GEMDObject(
                gemd_type=f["type"],
                uid=f["uids"][uuid][:64],
                context=json.dumps(f),
                gemdmodel=model,
            )
        except Exception as e:
            print("ERROR:", e)

    session = Session(db.ENGINE)
    session.add(model)
    session.commit()
    print("Session commited.")

    # print("Loading GEMDEdges and GEMDAttributes...")
    # # creating GEMDEdges
    # db.execute_query(create_edges_query(model.id))

    # # creating GEMDEdges
    # db.execute_query(create_attributes_query(model.id))

    print("Queries executed.")



"""Base class for processes and measurement."""

from typing import ClassVar, Type, Union, Optional
import re
from datetime import datetime

from gemd import (
    ProcessTemplate,
    ProcessSpec,
    ProcessRun,
    MeasurementTemplate,
    MeasurementSpec,
    MeasurementRun,
    Condition,
    Parameter,
    PerformedSource,
)

from .base_element import BaseElement
from openmsimodel.utilities.typing import SpecOrRunLiteral
from openmsimodel.utilities.attributes import finalize_template

__all__ = ["ProcessOrMeasurement"]


class ProcessOrMeasurement(BaseElement):
    """Base class for processes and measurements."""

    _TempType: ClassVar[Type[Union[ProcessTemplate, MeasurementTemplate]]]
    _SpecType: ClassVar[Type[Union[ProcessSpec, MeasurementSpec]]]
    _RunType: ClassVar[Type[Union[ProcessRun, MeasurementRun]]]

    TEMPLATE: ClassVar[Union[ProcessTemplate, MeasurementTemplate]]

    # {'brand1': {'model1': ('id1', 'id2', etc.), etc.}, etc.}
    # e.g. {'Quantum Design': {'DynaCool': ['1', '2', '3']}}
    INSTRUMENTS: ClassVar[dict[str, dict[str, tuple]]]

    # TODO: measurement specs dont take properties but runs do
    def __init__(
        self,
        name: str,
        *,
        template: ClassVar[Union[ProcessTemplate, MeasurementTemplate]] = None,
        notes: Optional[str] = None,
        conditions: Optional[list[Condition]] = None,
        parameters: Optional[list[Parameter]] = None,
        which: SpecOrRunLiteral = "spec",  # TODO: fix this
    ) -> None:
        super().__init__(name, template=template, notes=notes)

        if conditions is None:
            conditions = []

        if parameters is None:
            parameters = []

        # FIXME: would this work if template?
        self.update_conditions(
            *conditions, replace_all=True, which=which
        )  # TODO change which? or add param?

        self.update_parameters(*parameters, replace_all=True, which=which)

    def get_conditions_dict(self):
        """Return a ``dict`` of the spec and run conditions."""
        return self._spec_run_dict(self._spec.conditions, self._run.conditions)

    def update_conditions(
        self,
        *conditions: Condition,
        replace_all: bool = False,
        which: SpecOrRunLiteral = "spec",
    ) -> None:
        """
        Change or add conditions.

        Parameters
        ----------
        *conditions: Condition
            The conditions to change (by name) or add.
        replace_all: bool, default False
            If ``True``, remove any existing conditions before adding new ones.
        which: {'spec', 'run', 'both'}, default 'spec'
            Whether to update the spec, run, or both.

        Raises
        ------
        ValueError
            If the name of a condition is not supported.
        """

        self._update_attributes(
            AttrType=Condition,
            attributes=conditions,
            replace_all=replace_all,
            which=which,
        )

    def remove_conditions(
        self, *condition_names: str, which: SpecOrRunLiteral = "spec"
    ) -> None:
        """
        Remove conditions by name.

        *condition_names: str
            The names of conditions to remove.
        which: {'spec', 'run', 'both'}, default 'spec'
            Whether to remove from the spec, run, or both.

        Raises
        ------
        ValueError
            If the name of a condition is not supported.
        """

        self._remove_attributes(
            AttrType=Condition, attr_names=condition_names, which=which
        )

    def get_parameters_dict(self):
        """Return a ``dict`` of the spec and run parameters."""
        return self._spec_run_dict(self._spec.parameters, self._run.parameters)

    def update_parameters(
        self,
        *parameters: Parameter,
        replace_all: bool = False,
        which: SpecOrRunLiteral = "spec",
    ) -> None:
        """
        Change or add parameters.

        Parameters
        ----------
        *parameters: Parameter
            The parameters to change (by name) or add.
        replace_all: bool, default False
            If ``True``, remove any existing parameters before adding new ones.
        which: {'spec', 'run', 'both'}, default 'spec'
            Whether to update the spec, run, or both.

        Raises
        ------
        ValueError
            If the name of a parameter is not supported.
        """

        self._update_attributes(
            AttrType=Parameter,
            attributes=parameters,
            replace_all=replace_all,
            which=which,
        )

    def remove_parameters(
        self, *parameter_names: str, which: SpecOrRunLiteral = "spec"
    ) -> None:
        """
        Remove parameters by name.

        *parameter_names: str
            The names of parameters to remove.
        which: {'spec', 'run', 'both'}, default 'spec'
            Whether to remove from the spec, run, or both.

        Raises
        ------
        ValueError
            If the name of a parameter is not supported.
        """

        self._remove_attributes(
            AttrType=Parameter, attr_names=parameter_names, which=which
        )

    def get_source(self) -> dict[str, str]:
        """
        Get the run's source.

        Returns
        -------
        source: dict
            ``{'performed_by': '<email>', 'performed_date': '<iso_date>'}``
        """
        if not self._run.source:
            return
        run_source = self._run.source.as_dict()

        return {
            "performed_by": run_source["performed_by"],
            "performed_date": run_source["performed_date"],
        }

    def set_source(self, email: str, iso_date: Optional[str] = None) -> None:
        """
        Set the run's source with a valid email address and an optional ISO date string.

        Parameters
        ----------
        email: str
            A valid email address.
        date: str, optional
            A date string to be passed to ``datetime.fromisoformat``.

        Raises
        ------
        ValueError
            If `email` is invalid.
        """

        # if not re.fullmatch(r"[^@]+@[^@]+\.[^@]+", email):
        #     raise ValueError(
        #         'Invalid email. Must contain a single "@" and at least one "." after the "@".'
        #     )

        if iso_date is not None:
            iso_date = datetime.fromisoformat(iso_date).isoformat(timespec="auto")

        self._run.source = PerformedSource(email, iso_date)

    # TODO: Move this to base node?
    @staticmethod
    def _spec_run_dict(
        spec_attrs: Union[list[Condition], list[Parameter]],
        run_attrs: Union[list[Condition], list[Parameter]],
    ):
        """
        Return a ``dict`` of spec and run conditions or parameters.

        The keys are the names of the spec and run attributes.
        Each value is a ``dict`` with the keys ``'spec'`` and ``'run'``.
        Each ``'spec'`` and ``'run'`` key corresponds to another ``dict`` containing
        a value ``dict`` and origin ``str``, or ``None`` if the attribute is only
        found under one of spec or run.

        Example output:

        {
            'Duration': {
                'spec': {
                    'value': {
                        'nominal': 60,
                        'units': 'minutes',
                        'type': 'nominal_real'
                    },
                    'origin': 'specified'
                },
                'run': {
                    'value': None,
                    'origin': None
                }
            },
            'Set temperature': {
                'spec': {
                    'value': {
                        'nominal': 105,
                        'units': 'K',
                        'type': 'nominal_real'
                    },
                    'origin': 'specified'
                },
                'run': {
                    'value': {
                        'lower_bound': 104,
                        'upper_bound': 106,
                        'units': 'K',
                        'type': 'uniform_real'
                    },
                    'origin': 'measured'
                }
            },
            'Set pressure': {
                'spec': {
                    'value': {
                        'nominal': 400,
                        'units': 'millitorr',
                        'type': 'nominal_real'
                    },
                    'origin': 'specified'
                },
                'run': {
                    'value': {
                        'mean': 399.9,
                        'std': 0.5,
                        'units': 'K',
                        'type': 'normal_real'
                    },
                    'origin': 'measured'
                }
            }
        }
        """

        spec_dict = {
            attr.name: {"value": attr.value.as_dict(), "origin": attr.origin}
            for attr in spec_attrs
        }
        run_dict = {
            attr.name: {"value": attr.value.as_dict(), "origin": attr.origin}
            for attr in run_attrs
        }

        attrs = {}

        for name, value_and_origin in spec_dict.items():
            attrs[name] = {"spec": value_and_origin, "run": None}

        for name, value_and_origin in run_dict.items():
            if name in attrs:
                attrs[name]["run"] = value_and_origin
            else:
                attrs[name] = {"spec": None, "run": value_and_origin}

        return attrs

    def to_form(self) -> str:
        pass



'''Support for group attributes.'''

from typing import TypeAlias, ClassVar, Union, Optional, Any
import re

import numpy as np
import pandas as pd

from gemd import (
    MeasurementTemplate, MeasurementSpec, MeasurementRun,
    MaterialRun, Condition, Parameter, Property, FileLink,
    NominalCategorical, EmpiricalFormula, NominalInteger, NominalReal
)
from gemd.entity.template.attribute_template import AttributeTemplate
from gemd.entity.attribute.base_attribute import BaseAttribute

from .base_special_attrs import BaseSpecialAttrs
from .utils import AttrTypes, attr_template_dict, generate_tags, generate_source

__all__ = ['GrpAttrs']

Measurement: TypeAlias = Union[MeasurementSpec, MeasurementRun]

class GrpAttrs(BaseSpecialAttrs):
    '''
    Represent a single measurement's group attributes as ``DataFrame``s of
    attributes.

    Each row of the ``DataFrame``s represents a group of attributes. Each column
    is an attribute name.
    '''

    GRP: ClassVar[str] = '_grp'
    RE: ClassVar[re.Pattern] = re.compile(f'^(.+){GRP}(\\d+)$')
    CAT_STR: ClassVar[str] = 'categorical'
    FORM_STR: ClassVar[str] = 'formula'

    def __init__(self, measurement: Measurement) -> None:
        '''
        Parse a measurement's group attributes and create ``DataFrame``s for the
        conditions, parameters, and properties.

        Group attributes are expected to be named as ``<base_name>_grp<group>``.

        Parameters
        ----------
        measurement: MeasurmentSpec or MeasurementRun
            Measurement containing appropriately-named group attributes.
        '''        

        self._measurement = measurement

        conditions_tuple = self._parse(measurement.conditions)
        self._grp_conds: pd.DataFrame = conditions_tuple[0]
        self._other_conds: list[Condition] = conditions_tuple[1]

        parameters_tuple = self._parse(measurement.parameters)
        self._grp_params: pd.DataFrame = parameters_tuple[0]
        self._other_params: list[Parameter] = parameters_tuple[1]

        self._grp_props: pd.DataFrame = pd.DataFrame()
        self._other_props: list[Property] = []

        if isinstance(measurement, MeasurementRun):
            properties_tuple = self._parse(measurement.properties)
            self._grp_props = properties_tuple[0]
            self._other_props = properties_tuple[1]

        self._reindex_all_dfs()

    @property
    def measurement(self) -> Measurement:
        '''The measurement containing these attributes.'''
        return self._measurement

    @property
    def group_conditions(self) -> pd.DataFrame:
        '''Conditions of groups.'''
        return self._grp_conds
    
    @property
    def other_conditions(self) -> list[Condition]:
        '''Non-group conditions.'''
        return self._other_conds

    @property
    def group_parameters(self) -> pd.DataFrame:
        '''Parameters of groups.'''
        return self._grp_params
    
    @property
    def other_parameters(self) -> list[Parameter]:
        '''Non-group parameters.'''
        return self._other_params

    @property
    def group_properties(self) -> pd.DataFrame:
        '''Properties of groups.'''
        return self._grp_props
    
    @property
    def other_properties(self) -> list[Property]:
        '''Non-group properties.'''
        return self._other_props

    def concat(self, df: pd.DataFrame) -> None:
        '''
        Concatenate group attribute data.

        Indices in the new data are respected and will replace rows with
        identical indices in the existing data.

        Parameters
        ----------
        df: DataFrame
            Should contain the appropriate attribute columns to update the
            groups.
        '''

        attrs = pd.concat(self._df_list(), axis=1)
        attrs = pd.concat((attrs, df), axis=0)
        attrs = attrs[~attrs.index.duplicated(keep='last')]

        self._grp_conds = attrs.loc[:, self._grp_conds.columns]
        self._grp_params = attrs.loc[:, self._grp_params.columns]
        self._grp_props = attrs.loc[:, self._grp_props.columns]

        self._reindex_all_dfs()
        self._reindex_all_attrs()

    def remove(self, index: int) -> None:
        '''Remove a row from the attributes by index, then reindex.'''

        new_index = np.arange(len(self._grp_conds)-1)

        for df in self._df_list():
            df.drop(index=index, inplace=True)
            df.index = new_index

        self._reindex_all_attrs()

    def update_object(self) -> None:
        '''
        Update the ``MeasurementSpec`` or ``MeasurmentRun`` used to initialize
        with the attributes.

        This is performed in-place on the existing measurement and will
        overwrite any attributes added after ``GrpAttrs`` initialization.
        '''

        grp_conds = self._extract_attrs_from_df(self._grp_conds)
        grp_params = self._extract_attrs_from_df(self._grp_params)

        self._measurement.conditions = self._other_conds + grp_conds
        self._measurement.parameters = self._other_params + grp_params

        if isinstance(self._measurement, MeasurementRun):
            grp_props = self._extract_attrs_from_df(self._grp_props)
            self._measurement.properties = self._other_props + grp_props

    @classmethod
    def object_from_file(
        cls,
        filename: str,
        url: str,
        material_run: MaterialRun,
        measurement_template: MeasurementTemplate,
        attr_types: AttrTypes,
        columns: list[str],
        names: list[str],
        units: list[Union[str, None]],
        origins: list[str],
        brand: Optional[str] = None,
        model: Optional[str] = None,
        instr_id: Optional[str] = None,
        email: Optional[str] = None,
        iso_date: Optional[str] = None,
        **read_csv_kwargs: Any
        ) -> tuple[MeasurementSpec, MeasurementRun]:
        '''
        Create an empty ``MeasurementSpec`` and filled ``MeasurementRun`` from a delimited file.

        Parameters
        ----------
        filename: str
            The name of the file.
        url: str
            The URL or file path leading to `filename`.
        material_run: str
            The relevant ``MaterialRun``.
        measurement_template: MeasurementTemplate
            The measurement template for this measurement.
        attr_types: list of Type[Condition], Type[Parameter], Type[Property]
            The type of each attribute in the data, in order.
        columns: list[str]
            The names of the data columns.
        names: list[str]
            The names of the attributes corresponding to an attribute in `measurement_template`.
        units: list[str or None]
            The units of the attribute values. ``"categorical"`` for
            NominalCategorical, ``"formula"``, for EmpiricalFormula,
            other ``str`` for NominalReal, ``None`` for NominalInteger.
        origins: list[str]
            The origins of the attribute values.
        brand: str, optional
            The brand of the instrument used for the measurement.
        model: str, optional
            The model of the instrument used for the measurement.
            Ignored if `brand` is ``None``.
        instr_id: str, optional
            A further identifier for the instrument used, such as a number.
            Ignored if `brand` or `model` is ``None``.
        email: str, optional
            Email of the person who performed the measurement.
        iso_date: str, optional
            Date the measurement was performed, in ISO format.
        **read_csv_kwargs: Any
            Passed to ``pandas.read_csv()``.
        '''

        df = pd.read_csv(filename, **read_csv_kwargs)

        tags = generate_tags(brand, model, instr_id)

        source = generate_source(email, iso_date)

        template_dict = attr_template_dict(measurement_template)

        if not (len(attr_types) == len(columns) == len(names) == len(units) == len(origins)):
            raise ValueError('Iterable inputs must be equal in length.')

        conditions, parameters, properties = cls._create_attrs_from_df(
            df, template_dict, attr_types, columns, names, units, origins
        )

        spec = MeasurementSpec(
            name=measurement_template.name,
            template=measurement_template,
            tags=tags
        )

        run = MeasurementRun(
            name=measurement_template.name,
            spec=spec,
            tags=tags,
            material=material_run,
            source=source,
            conditions=conditions,
            parameters=parameters,
            properties=properties,
            file_links=[FileLink(filename=filename, url=url)]
        )

        return spec, run

    @classmethod
    def _parse(cls, attrs: list[BaseAttribute]) -> tuple[pd.DataFrame, list[BaseAttribute]]:
        '''
        Match each attribute's name with regular expression.

        Base name determines the column. _grp value determines the row.
        '''

        grp_df = pd.DataFrame()
        other_attrs = []

        for attr in attrs:

            match = cls.RE.match(attr.name)

            if match is not None:
                cls._add_attr_to_df(attr, match, grp_df)
            else:
                other_attrs.append(attr)

        reindexed_df = cls._reindex_df(grp_df)

        return reindexed_df, other_attrs

    @classmethod
    def _add_attr_to_df(
        cls,
        attr: BaseAttribute,
        match: re.Match[str],
        grp_df: pd.DataFrame
        ) -> None:
        '''Given a matching attribute, add it to `grp_df`.'''

        # regular expression groups
        match_groups = match.groups()
        base_name = match_groups[0]
        grp_val = int(match_groups[1])
  
        # add column if not there
        if base_name not in grp_df.columns:
            grp_df[base_name] = np.nan

        # place attribute in appropriate df
        grp_df.loc[grp_val, base_name] = attr

    @classmethod
    def _reindex_df(cls, grp_df: pd.DataFrame) -> pd.DataFrame:
        '''Add in any missing integer indices and sort.'''

        if len(grp_df) > 0:
            grp_df = grp_df.reindex(np.arange(grp_df.index.max()+1))

        return grp_df

    def _reindex_all_dfs(self) -> None:
        '''Ensure that each ``DataFrame`` has the same (maximum) number of groups.'''

        if len(self._grp_conds) == len(self._grp_params):
            return

        grp_max = int(np.nanmax([self._grp_conds.index.max(), self._grp_params.index.max()]))

        self._grp_conds = self._grp_conds.reindex(np.arange(grp_max+1))
        self._grp_params = self._grp_params.reindex(np.arange(grp_max+1))
        self._grp_props = self._grp_props.reindex(np.arange(grp_max+1))

    def _df_list(self) -> list[pd.DataFrame]:
        '''Return a list of the attributes ``DataFrame``s.'''
        return [self._grp_conds, self._grp_params, self._grp_props]

    def _reindex_all_attrs(self) -> None:
        '''Rename all attributes according to where they are in their ``DataFrame``.'''

        for df in self._df_list():
            for name, series in df.items():
                for i, attr in series.items():
                    if not pd.isna(attr):
                        attr.name = f'{name}{self.GRP}{i}'

    @classmethod
    def _create_attrs_from_df(
        cls,
        df: pd.DataFrame,
        template_dict: dict[str, AttributeTemplate],
        attr_types: AttrTypes,
        columns: list[str],
        names: list[str],
        units: list[Union[str, None]],
        origins: list[str]
        ) -> tuple[list[Condition], list[Parameter], list[Property]]:
        '''Create conditions, parameters, and properties from the `df` values.'''

        conditions = []
        parameters = []
        properties = []

        for a, c, n, u, o in zip(attr_types, columns, names, units, origins):

            if a is Condition:
                to_append = conditions
            elif a is Parameter:
                to_append = parameters
            elif a is Property:
                to_append = properties
            else:
                raise ValueError(
                    'attr_types must consist of Type[Condition], Type[Parameter], '
                    f'or Type[Property], but found {type(a)}.'
                )

            for i, val in enumerate(df.loc[:, c]):

                if u == cls.CAT_STR:
                    value = NominalCategorical(val)
                elif u == cls.FORM_STR:
                    value = EmpiricalFormula(val)
                elif u is None:
                    value = NominalInteger(val)
                else:
                    value = NominalReal(val, u)

                to_append.append(a(
                    name=f'{n}{cls.GRP}{i}',
                    value=value,
                    template=template_dict[n],
                    origin=o
                ))

        return conditions, parameters, properties



from typing import ClassVar

from gemd import (
    MeasurementTemplate,
    PropertyTemplate,
    CategoricalBounds,
    NominalCategorical,
    RealBounds,
    NominalReal,
)
from openmsimodel.entity.base.measurement import Measurement
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
    _validate_temp_keys,
)


class SRJT(Measurement):
    """Class representing SRJT as a measurement"""

    TEMPLATE: ClassVar[MeasurementTemplate] = MeasurementTemplate(
        name="SRJT", description="Strain rate jump measurement"
    )

    _ATTRS: ClassVar[AttrsDict] = _validate_temp_keys(TEMPLATE)

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Avg. Corrected Hardness (Gpa)", bounds=RealBounds(0, 5, "GPa")
        ),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Corrected Hardness Std. Dev.", bounds=RealBounds(0, 1, "")
        ),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Corrected Modulus Std. Dev.", bounds=RealBounds(0, 1, "")
        ),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Avg. Corrected Modulus (Gpa)", bounds=RealBounds(0, 400, "GPa")
        ),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Avg. Pile-up Ratio", bounds=RealBounds(0, 200, "")
        ),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Avg. Raw Modulus (Gpa)", bounds=RealBounds(0, 500, "GPa")
        ),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Avg. Raw Hardness (Gpa)", bounds=RealBounds(0, 5, "GPa")
        ),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Strain Rate Sensitivity Exponent", bounds=RealBounds(0, 1, "")
        ),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Strain Rate Sensitivity Exponent Std. Dev.",
            bounds=RealBounds(0, 1, ""),
        ),
    )

    finalize_template(_ATTRS, TEMPLATE)



from typing import ClassVar

from gemd import (
    MaterialTemplate,
    PropertyTemplate,
    CategoricalBounds,
    NominalCategorical,
)
from gemd.entity.bounds import RealBounds
from gemd.entity.value import NominalReal

from openmsimodel.entity.base.material import Material
from openmsimodel.utilities.attributes import AttrsDict, define_attribute, finalize_template

__all__ = ["Element"]


class Element(Material):
    """Class representing element"""

    TEMPLATE: ClassVar[MaterialTemplate] = MaterialTemplate(
        "Element",
        description="Individual ingredient of a composition",
    )
    
    _ATTRS: ClassVar[AttrsDict] = {"properties": {}}
    
    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            'Composition Percentage',
            bounds=RealBounds(0,100,''),
            description='Percentage of the total composition that the material takes in the overall composition'
        ),
        default_value=NominalReal(0.0,'')
    )
    
    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            'Target Mass',
            bounds=RealBounds(0,150,'g'),
            description='Expected mass to attain based on composition design '
        ),
        default_value=NominalReal(0.0,'g')
    )
    
    

    finalize_template(_ATTRS, TEMPLATE)



from openmsimodel.entity.base.material import Material
from openmsimodel.entity.base.process import Process
from openmsimodel.entity.base.ingredient import Ingredient
from openmsimodel.entity.base.measurement import Measurement

from typing import ClassVar, Type, Optional

from openmsimodel.utilities.typing import (
    Spec,
    Run,
)


def from_spec_or_run(
    name: str,
    *,
    notes: Optional[str] = None,
    spec: Spec = None,
    run: Run = None,
) -> "BaseElement":
    """
    Instantiate a `BaseElement` from a spec or run with appropriate validation.

    Note that the spec's and run's name and notes will be set to `name` and
    `notes`, the spec's template will be set to the class template,
    and the run's spec will be set to this spec.
    """
    if spec is None and run is None:
        raise ValueError("At least one of spec or run must be given.")

    if (spec is not None and spec.typ.startswith("material")) or (
        run is not None and run.typ.startswith("material")
    ):
        return Material.from_spec_or_run(name=name, notes=notes, spec=spec, run=run)
    elif (spec is not None and spec.typ.startswith("process")) or (
        run is not None and run.typ.startswith("process")
    ):
        return Process.from_spec_or_run(name=name, notes=notes, spec=spec, run=run)
    elif (spec is not None and spec.typ.startswith("measurement")) or (
        run is not None and run.typ.startswith("measurement")
    ):
        return Measurement.from_spec_or_run(name=name, notes=notes, spec=spec, run=run)
    elif (spec is not None and spec.typ.startswith("ingredient")) or (
        run is not None and run.typ.startswith("ingredient")
    ):
        return Ingredient.from_spec_or_run(name=name, notes=notes, spec=spec, run=run)



import json
import networkx as nx
from collections import defaultdict
import shutil
import os
from gemd.util.impl import recursive_foreach
from gemd.json import GEMDJson

import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import matplotlib.colors as colors
import matplotlib.cm as cmx
import matplotlib.patches as mpatches
from matplotlib import pylab
from pathlib import Path

# _encoder = None
# _subdirpath = None


def out(item):
    """
    function object to run on individual item during recursion
    :param item: json item to write its destination
    """
    fn = "_".join([item.__class__.__name__, item.name, item.uids["auto"], ".json"])
    with open(os.path.join(_subdirpath, fn), "w") as fp:
        fp.write(_encoder.thin_dumps(item, indent=3))


def handle_value(G, uid, attributes):
    # TODO: add pointing to templates?
    for att in attributes:
        if att["type"] == "property_and_conditions":
            value = att["property"]["value"]
            att_name = att["property"]["name"]
        else:
            value = att["value"]
            att_name = att["name"]

        if value["type"] == "nominal_real":
            node_name = "{}, {} {}".format(att_name, value["nominal"], value["units"])
            G.add_node(node_name, shape="rectangle", color="orange")
            G.add_edge(uid, node_name)
            # G.add_edge(node_name, uid)
        if value["type"] == "nominal_categorical":
            node_name = "{}, {}".format(att_name, value["category"])
            G.add_node(node_name, shape="rectangle", color="orange")
            # G.add_edge(node_name, uid)
            G.add_edge(uid, node_name)


def handle_attributes(G, uid, obj_data, object_class, obj_state):
    if "parameters" in obj_data:
        handle_value(G, uid, obj_data["parameters"])
    if "properties" in obj_data:
        handle_value(G, uid, obj_data["properties"])
    if "conditions" in obj_data:
        handle_value(G, uid, obj_data["conditions"])


def handle_gemd_obj(G, uid, obj_data, obj_type, obj_state, add_attributes):
    if obj_type.startswith("process"):
        if obj_type.endswith(obj_state):
            G.add_node(uid, color="red")
            if add_attributes:
                handle_attributes(G, uid, obj_data, "process", obj_state)
    elif obj_type.startswith("ingredient"):  # TODO if node doesn't exist, create?
        if obj_type.endswith(obj_state):
            G.add_node(uid, color="blue")
            process = obj_data["process"]["id"]
            G.add_edge(uid, process)
            if add_attributes:
                handle_attributes(G, uid, obj_data, "ingredient", obj_state)
            if "material" in obj_data and obj_data["material"]:
                material = obj_data["material"]["id"]
                G.add_edge(material, uid)
    elif obj_type.startswith("material"):
        if obj_type.endswith(obj_state):
            G.add_node(uid, color="green")
            if add_attributes:
                handle_attributes(G, uid, obj_data, "material", obj_state)
            # if "process" in obj_data and obj_data["process"]:
            if obj_data["process"] and obj_data["process"]:
                process = obj_data["process"]["id"]
                G.add_edge(process, uid)  # ?
    elif obj_type.startswith("measurement"):
        if obj_type.endswith(obj_state):
            G.add_node(uid, color="purple")
            if add_attributes:
                handle_attributes(G, uid, obj_data, "measurement", obj_state)
            if "material" in obj_data and obj_data["material"]:
                material = obj_data["material"]["id"]
                G.add_edge(uid, material)


def save_graph(graph, file_name):
    plt.figure(num=None, figsize=(20, 20), dpi=80)
    plt.axis("off")
    fig = plt.figure(1)
    pos = nx.spring_layout(graph)
    nx.draw_networkx_nodes(graph, pos)
    nx.draw_networkx_edges(graph, pos)
    nx.draw_networkx_labels(graph, pos)

    cut = 1.00
    xmax = cut * max(xx for xx, yy in pos.values())
    ymax = cut * max(yy for xx, yy in pos.values())
    plt.xlim(0, xmax)
    plt.ylim(0, ymax)

    plt.savefig(file_name, bbox_inches="tight")
    pylab.close()
    del fig


"""helper to extract GEMD data from all scenarios, whether folder of JSONs or single JSON, thin or full, etc.
    it raises IOError in case the data can't be properly extracted.
    """


def read_gemd_data(dirpath, encoder):
    """helper to extract GEMD data from all scenarios, whether folder of JSONs or single JSON, thin or full, etc.
    it raises IOError in case the data can't be properly extracted.

    Args:
        dirpath (str, Path): path to directory or file containing GEMD knowledge
        encoder (GEMDJson): GEMD encoder

    Raises:
        IOError: if folder or file doesn't match the criteria

    Returns:
        list: all gemd objects extracted from folder or file
    """
    gemd_objects = []
    gemd_paths = []
    # try:
    if True:
        if type(dirpath) == list:
            print("Extracting list...")
            for obj in dirpath:
                gemd_objects.append(json.loads(encoder.thin_dumps(obj)))
        elif os.path.isdir(dirpath):
            print("Extracting folder...")
            for dp, dn, filenames in os.walk(dirpath):
                for f in filenames:
                    if f.endswith(".json"):
                        path = os.path.join(dp, f)
                        with open(path) as fp:
                            gemd_objects.append(json.load(fp))
                            gemd_paths.append(Path(path))
        elif os.path.isfile(dirpath) and str(dirpath).endswith(".json"):
            print("Extracting file...")
            with open(dirpath) as fp:
                gemd_objects = json.load(fp)

            #     gemd_objects = json.load(fp)
    # except:
    #     raise IOError(  # FIXME
    #         f"couldn't extract GEMD data. Expected folder of JSONs or single JSON with 1+ objects. "
    #     )
    if len(gemd_objects) == 0:
        raise ValueError(f"Couldn't extract any gemd object from {dirpath}")
    return gemd_objects, gemd_paths


def plot_graph(
    dirpath, obj_state="run", objectpath=None, tmp="tmp", add_attributes=False
):
    """
    creates a NetworkX graph representation of the GEMD relationships by reading every object
    generated by the GEMDEncoder object, storing all of its links by uid, and forming directed relationships,
    such as ingredient->process, or process->material
    It then allows filtering the objects mapped (i.e., removing spec or runs,
    measurements or ingredients) and saves a NetworkX graph in "dot" as .png

    :param dirpath: source of graph
    :param obj_state: to plot a graph of specs, runs or templates
    """
    G = nx.DiGraph()
    object_mapping = defaultdict()
    name_mapping = defaultdict()
    encoder = GEMDJson()
    nb_disregarded = 0

    if objectpath:  # plotting a graph from a single, full json (!= thin)
        with open(objectpath) as fp:
            object = encoder.load(fp)
    else:  # plotting a graph from a bunch of thin jsons
        gemd_objects = [
            os.path.join(dp, f)
            for dp, dn, filenames in os.walk(dirpath)
            for f in filenames
            if f.endswith(".json")
        ]
    if len(gemd_objects) == 0:
        return

    # adding objects to graph one by one
    for i, obj in enumerate(gemd_objects):
        if "raw_jsons" in obj:
            continue
        fp = open(obj, "r")
        obj_data = json.load(fp)
        obj_type = obj_data["type"]
        if (
            obj_type.startswith("parameter")
            or obj_type.startswith("condition")
            or obj_type.startswith("property")
        ):
            nb_disregarded += 1
            continue
        obj_uid = obj_data["uids"]["auto"]
        obj_name = obj_data["name"]
        name_mapping[obj_uid] = "{},  {}".format(obj_name, obj_uid[:3])
        handle_gemd_obj(G, obj_uid, obj_data, obj_type, obj_state, add_attributes)
        if i % 1000 == 0:
            print("{} gemd objects processed...".format(i))

    # converting to grapviz
    _relabeled_G = nx.relabel_nodes(G, name_mapping)

    relabeled_G = nx.nx_agraph.to_agraph(_relabeled_G)
    relabeled_G.node_attr.update(nodesep=0.4)
    relabeled_G.node_attr.update(ranksep=1)

    relabeled_G.layout(prog="dot")

    # plotting
    relabeled_G.draw(os.path.join(dirpath, "{}_graph.svg".format(obj_state)))
    plt.close()
    with open(os.path.join(dirpath, "{}_graph.dot".format(obj_state)), "x") as f:
        f.write(str(relabeled_G))

    # info
    print("cycles in the graph: {}".format(list(nx.simple_cycles(G))))
    print(
        "nb of disregarded elements (i.e., templates/specs): {}/{}".format(
            nb_disregarded, len(gemd_objects)
        )
    )



from typing import ClassVar, Type, Optional
from openmsimodel.workflow.workflow import Workflow
from openmsimodel.entity.base.material import Material
from openmsimodel.entity.base.base_element import BaseElement
from openmsimodel.utilities.runnable import Runnable


class Subworkflow(Runnable):
    """Subworkflow represents 1+ BaseElements that are grouped together for easier instantiation, manipulation, analysis, and more.
    The grouping can be based on any rule or criteria of interest, such as consecutive elements, elements with a logical relationships or simply preference.
    """

    def __init__(
        self, name: str, *, workflow: Workflow = None, self_assign: bool = True
    ):
        """initialization of Subworkflow.

        Args:
            name (str): name of subworkflow
            workflow (Workflow, optional): the workflow that subworkflow belongs to. Defaults to None.
            self_assign (bool, optional): whether to add the current subworkflow to parent workflow's dictionary 'subs'. Defaults to True.
        """
        self.name = name
        self.workflow = workflow
        if self_assign:
            self.workflow.subs[name] = self

    @classmethod
    def from_(self, any):
        """returns a subworkflow instantiated from elements contained it. will vary based on subworkflow and element types."""
        pass

    @classmethod
    def get_command_line_arguments(cls):
        superargs, superkwargs = super().get_command_line_arguments()
        args = []
        kwargs = {**superkwargs}
        return args, kwargs

    @classmethod
    def run_from_command_line(cls, args=None):
        parser = cls.get_argument_parser()
        args = parser.parse_args(args=args)
        Subworkflow = cls(*args)


def main(args=None):
    """
    Main method to run from command line
    """
    Subworkflow.run_from_command_line(args)


if __name__ == "__main__":
    main()



from typing import ClassVar, Optional

from gemd import (
    MeasurementTemplate,
    MeasurementSpec,
    MeasurementRun,
    MaterialRun,
    Condition,
    Parameter,
    Property,
)
from gemd.entity.util import make_instance

# from openmsimodel.utilities.typing import ValueAndOrigin

from .process_or_measurement import ProcessOrMeasurement

__all__ = ["Measurement"]


class Measurement(ProcessOrMeasurement):
    """
    Base element for measurements.

    """

    _TempType = MeasurementTemplate
    _SpecType = MeasurementSpec
    _RunType = MeasurementRun

    TEMPLATE: ClassVar[MeasurementTemplate]

    def __init__(
        self,
        name: str,
        *,
        notes: Optional[str] = None,
        conditions: Optional[list[Condition]] = None,
        parameters: Optional[list[Parameter]] = None,
        properties: Optional[list[Property]] = None,
        material: Optional[MaterialRun] = None
    ) -> None:
        super().__init__(
            name, notes=notes, conditions=conditions, parameters=parameters
        )

        if properties is None:
            properties = []

        self.update_properties(properties=properties, replace_all=True)

        self.set_material(material)

    @property
    def spec(self) -> MeasurementSpec:
        """The underlying measurement spec."""
        return self._spec

    @property
    def run(self) -> MeasurementRun:
        """The underlying measurement run."""
        return self._run

    @classmethod
    def from_spec_or_run(
        cls,
        name: str,
        *,
        notes: Optional[str] = None,
        spec: MeasurementSpec = None,
        run: MeasurementRun = None
    ) -> "Measurement":
        """
        Instantiate a `Measurement` from a spec or run with appropriate validation.

        Note that the spec's template will be set to the class template,
        and the run's spec will be set to this spec.
        """

        if spec is None and run is None:
            raise ValueError("At least one of spec or run must be given.")

        measurement = cls(name, notes=notes)

        if spec is not None:
            if not isinstance(spec, MeasurementSpec):
                raise TypeError("spec must be a MeasurementSpec.")

            measurement._spec = spec

            measurement.spec.name = name
            measurement.spec.notes = notes
            measurement.spec.template = cls.TEMPLATE

            measurement.update_conditions(which="spec")
            measurement.update_parameters(which="spec")

        if run is not None:
            if not isinstance(run, MeasurementRun):
                raise TypeError("run must be a MeasurementRun.")

            measurement._run = run

            measurement.run.name = name
            measurement.run.notes = notes
            measurement.run.spec = measurement.spec

            measurement.update_conditions(which="run")
            measurement.update_parameters(which="run")
            measurement.update_properties()

            source = measurement.get_source()
            measurement.set_source(
                email=source["performed_by"], iso_date=source["performed_date"]
            )

        else:
            measurement._run = make_instance(measurement.spec)

        return measurement

    def get_material(self) -> str:
        """
        Get the name of the material on sstate the measurement run was performed.


        """

        return self._run.material.name

    def set_material(self, material: Optional[MaterialRun]) -> None:
        """
        Set the material on sstate the measurement run was performed.

        :param material: {MaterialRun, None}
            The measurement's material.
        """

        self._run.material = material

    # TODO: merge all properties stuff to base node?
    def get_properties_dict(self):
        """
        Return a ``dict`` of measurement run properties.
        The keys are the names of the properties.
        Each value is a ``dict`` containing a value ``dict`` and origin ``str``.
        """

        return self._prop_dict(self._run.properties)

    def update_properties(
        self, properties: Property, replace_all: bool = False
    ) -> None:
        """
        Change or add measured properties of the measurement run.
        properties: Property
        The properties to change (by name) or add.
        replace_all: bool, default False
        If ``True``, remove any existing properties before adding new ones.
        """

        self._update_attributes(
            AttrType=Property,
            attributes=properties,
            replace_all=replace_all,
            which="run",
        )

    def remove_properties(self, *property_names: str) -> None:
        """
        Remove measured properties from the measurement run by name.

        property_names: str
            The names of properties to remove.


        """

        self._remove_attributes(
            AttrType=Property, attr_names=property_names, which="run"
        )

    @staticmethod
    def _prop_dict(run_props: list[Property]):
        """Return a ``dict`` of measurement run properties."""

        prop_dict = {
            prop.name: {"value": prop.value.as_dict(), "origin": prop.origin}
            for prop in run_props
        }

        return prop_dict

    def to_form(self) -> str:
        pass



from typing import ClassVar

from gemd import (
    MeasurementTemplate,
    PropertyTemplate,
    CategoricalBounds,
    NominalCategorical,
    RealBounds,
    NominalReal,
)
from openmsimodel.entity.base.measurement import Measurement
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
)

__all__ = ["XRD"]


class XRD(Measurement):
    """Class representing XRD as a measurement"""

    TEMPLATE: ClassVar[MeasurementTemplate] = MeasurementTemplate(
        name="XRD", description="XRD measurement"
    )

    _ATTRS: ClassVar[AttrsDict] = {"conditions": {}, "parameters": {}, "properties": {}}

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Lattice Parameter", bounds=RealBounds(0, 5, "cm")
        ),
        # default_value=NominalReal(0, "cm"),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Phase", bounds=CategoricalBounds(categories=["FCC"])
        ),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Hardness, HV", bounds=RealBounds(0, 1000, "Pa")
        ),
        # default_value=NominalReal(0, "Pa"),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(name="SD, HV", bounds=RealBounds(0, 1000, "Pa")),
        # default_value=NominalReal(0, "Pa"),
    )

    finalize_template(_ATTRS, TEMPLATE)



import setuptools

version = "0.0.1"

setupkwargs = dict(
    name="openmsimodel",
    version=version,
    description="data modelling package",
    url="https://github.com/openmsi/openmsimodel",
    author="Ali Rachidi",
    author_email="arachid1@jhu.edu",
    license="MIT",
    packages=setuptools.find_packages(include=["openmsimodel*"]),
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "open_db=openmsimodel.db.open_db:main",
            "open_graph=openmsimodel.graph.open_graph:main",
        ],
    },
    python_requires=">=3.7,<3.10",
    install_requires=[
        "gemd==1.13.0",
        "networkx",
        "pymssql==2.2.8",
        "SQLAlchemy==2.0.17",
        "PyInquirer",
        "prompt_toolkit",
        "numpy",
        "pandas",
        "methodtools",
        "matplotlib",
        "yfiles_jupyter_graphs",
        "ipython",
        "ipykernel",
        "openpyxl",
        "graphviz",
        "questionary",
    ],
    extras_require={},
)

setupkwargs["extras_require"]["all"] = sum(setupkwargs["extras_require"].values(), [])

setuptools.setup(**setupkwargs)



"""Process and measurement classes."""

# from . import base
# from . import materials
# from . import processes
# from . import measurements

# from .base import *
# from .materials import *
# from .processes import *
# from .measurements import *

# __all__ = base.__all__ + materials.__all__ + processes.__all__ + measurements.__all__

# del materials
# del processes
# del measurements
# del base



"""
OpenDB - Database Interaction Tool
==================================

OpenDB is a command-line tool that allows interaction with a database for managing model artifacts. It provides capabilities for loading models, executing queries, and more.

Usage:
    python open_db.py [options]

Options:
    --database_name=<name>       Name of the database to connect to.
    --private_path=<path>       Path to a JSON file containing database credentials.
    --output=<path>             Path to the output directory.

Example:
    python open_db.py --database_name=GEMD_DB --private_path=credentials.json --output=output_dir

"""

import json
import random
import time
import os
from pathlib import Path

from questionary import prompt, select, text, confirm

from openmsimodel.db.gemd_database import MSSQLDatabase
from openmsimodel.utilities.argument_parsing import OpenMSIModelParser
from openmsimodel.utilities.runnable import Runnable
from openmsimodel.utilities.logging import Logger
import openmsimodel.db.queries as queries

from inspect import getmembers, isfunction


def create_acronym(phrase):
    # Function to create an acronym from a phrase
    acronym = "".join(word[0].upper() for word in phrase.split("_"))
    return acronym


class OpenDB(Runnable):
    """
    OpenDB - Database Interaction Tool

    This tool allows interaction with a database for managing model artifacts. It provides capabilities for loading models, executing queries, and more.
    """

    ARGUMENT_PARSER_TYPE = OpenMSIModelParser

    def __init__(self, database_name, private_path, output):
        """
        Initialization of OpenDB instance.

        :param database_name: Name of the database to connect to.
        :type database_name: str
        :param private_path: Path to a JSON file containing database credentials.
        :type private_path: str
        :param output: Path to the output directory.
        :type output: str
        """
        self.auth = None
        self.gemd_db = None
        self.output = output
        self.listed_functions = {}
        self.listed_acronyms = {}
        self.logger = Logger()

        self.setup(database_name, private_path)

    def setup(self, database_name, private_path):
        """
        Set up the OpenDB tool.

        :param database_name: Name of the database to connect to.
        :param private_path: Path to a JSON file containing database credentials.
        """
        try:
            with open(private_path, "r") as f:
                self.auth = json.load(f)
            self.database = database_name  # GEMD
            self.gemd_db = MSSQLDatabase(self.auth, self.database)
        except FileNotFoundError as e:
            self.logger.error(f"Error loading credentials: {e}")
            self.logger.warning("Database functionality disabled.")
        try:
            for function_name, function in getmembers(queries, isfunction):
                acc = create_acronym(function_name)
                self.listed_acronyms[acc] = function_name
                self.listed_functions[function_name] = (function, acc)
        except Exception as e:
            print(f"Error while reading functions from 'queries': {e}")

    def load_model(self, name, dirpath, uuid="auto"):
        return queries.load_model_query(name, self.gemd_db, dirpath, uuid)

    def create_tables(self):
        return queries.create_tables_query()

    def record_query_results(self, sql_results, query, name, dump=True):
        """
        Print query results and optionally save them to a CSV file.

        :param sql_results: The results of the SQL query.
        :type sql_results: DataFrame
        :param query: The SQL query that produced the results.
        :type query: str
        :param name: A name to use for the output file.
        :type name: str
        :param dump: If True, save the results to a CSV file. Default is True.
        :type dump: bool

        :return: The input DataFrame (sql_results).
        :rtype: DataFrame

        :raises ValueError: If an error occurs while saving to the output file.

        :examples:
        
        Example 1:
        
        .. code-block:: python
        
            result_df = self.load_model("example_model", "/path/to/model")
            query = "SELECT * FROM data"
            self.print_and_dump(result_df, query, "data_output")
        """
        print(f"Query: {query}")

        timestamp = time.strftime("%m%d%Y_%H%M_", time.localtime())
        random_suffix = random.randint(0, 10000)
        output_file = os.path.join(
            self.output, f"{name}_{timestamp}{random_suffix}.csv"
        )

        print(sql_results)

        try:
            if dump:
                sql_results.to_csv(output_file, index=False)
                print(f"Results saved to '{output_file}'")
            else:
                print("Results not saved to a file.")
        except Exception as e:
            error_message = f"Error saving results to '{output_file}': {e}"
            print(error_message)
            raise ValueError(error_message)

        return sql_results

    def list_queries(self):
        """
        List and display available queries along with their documentation.
        """
        for function_name, (
            function,
            function_acronym,
        ) in self.listed_functions.items():
            function_doc = function.__doc__ or "No documentation available"
            function_doc = function_doc.strip()

            max_doc_length = 240
            if len(function_doc) > max_doc_length:
                function_doc = function_doc[:max_doc_length] + " ..."

            print(f"- Name: {function_name}")
            print(f"  Acronym: {function_acronym}")
            print(f"  Description: {function_doc}")

    def interactive_mode(self):
        try:
            while True:
                # Present the user with a list of interactive modes to choose from
                mode_choices = [
                    "Load Model",
                    "Listed Queries",
                    "Custom Query",
                    "Add Schema",
                ]
                mode_question = select(
                    "Select an interactive mode:", choices=mode_choices
                ).ask()

                if mode_question == "Add Schema":
                    try:
                        # Create tables in the database schema
                        add_schema_query = queries.create_tables_query()
                        result = self.gemd_db.execute_query(add_schema_query)
                        self.record_query_results(
                            result, add_schema_query, "add_schema_query"
                        )
                    except Exception as e:
                        print(f"ERROR: {e}")
                        print("Try again.")

                elif mode_question == "Load Model":
                    # Input parameters for loading a model
                    params = input(
                        "Enter the appropriate parameters for loading a model (name and dirpath): "
                    )
                    params = params.split(" ")
                    if len(params) != 2:
                        raise ValueError("Expected two parameters (name and dirpath).")
                    self.load_model(params[0], params[1])

                elif mode_question == "Listed Queries":
                    # Display documentation on available queries
                    display_documentation = confirm(
                        "Would you like to display documentation on listed queries?"
                    ).ask()

                    if display_documentation:
                        self.list_queries()

                    try:
                        # Select a query and specify its required arguments
                        selected_query = input(
                            "Enter query with its required arguments: "
                        )
                        selected_query = selected_query.split(" ")
                        identifier = selected_query[0]

                        if (
                            identifier in self.listed_acronyms.keys()
                            or identifier.upper() in self.listed_acronyms.keys()
                        ):
                            name = self.listed_acronyms[identifier.upper()]
                            func = self.listed_functions[name][0]
                        elif (
                            identifier in self.listed_functions.keys()
                            or identifier.lower() in self.listed_functions.keys()
                        ):
                            func = self.listed_functions[identifier.lower()][0]
                        else:
                            raise KeyError(
                                f"Couldn't recognize the identifier passed as '{identifier}'. Pass the full name or acronym."
                            )

                        # Execute the selected query
                        selected_query = func(*selected_query[1:])
                        result = self.gemd_db.execute_query(selected_query)
                        self.record_query_results(result, selected_query, identifier)
                    except Exception as e:
                        print(f"ERROR: {e}")
                        print("Try again.")

                elif mode_question == "Custom Query":
                    # Enter a custom SQL query
                    custom_query = select("Enter a custom query: ").ask()
                    self.logger.info("Executing custom query...")
                    self.logger.info(f"query: {custom_query}")

                    try:
                        result = self.gemd_db.execute_query(custom_query)
                        self.record_query_results(result, custom_query, "custom_query")
                    except Exception as e:
                        print(f"ERROR: {e}")
                        print("Try again.")
                    self.logger.info("Done.")
        except KeyboardInterrupt:
            pass

    @classmethod
    def get_command_line_arguments(cls):
        superargs, superkwargs = super().get_command_line_arguments()
        args = [*superargs, "database_name", "private_path", "output"]
        kwargs = {**superkwargs}
        return args, kwargs

    @classmethod
    def run_from_command_line(cls, args=None):
        parser = cls.get_argument_parser()
        args = parser.parse_args(args=args)
        open_db = cls(args.database_name, args.private_path, args.output)
        open_db.interactive_mode()


def main(args=None):
    """
    Main method to run from command line
    """
    OpenDB.run_from_command_line(args)


if __name__ == "__main__":
    main()



from typing import ClassVar

from gemd import MaterialTemplate
from openmsimodel.entity.base.material import Material
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
)

__all__ = ["Traveler"]


class Traveler(Material):
    """Class representing a traveler"""

    TEMPLATE: ClassVar[MaterialTemplate] = MaterialTemplate(
        name="Traveler", description="Traveler"
    )

    _ATTRS: ClassVar[AttrsDict] = {"properties": {}}

    # define_attribute(
    #     _ATTRS,
    #     template=ParameterTemplate(
    #         name='Supplier',
    #         bounds=CategoricalBounds(categories=[''])
    #     ),
    #     default_value=NominalCategorical('')
    # )

    finalize_template(_ATTRS, TEMPLATE)
